# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This file contains the definition of the symbolic representation of 
parameters that can be used in Graphs. Var defines a variable, which can
either be a Boolean or continuous variable. Term is a product of variables,
and Poly is a sum of terms. 
Lark is used to define a parser that can translate a string into a Poly.
"""

from typing import Any, Callable, Union, Optional, Dict, List, Tuple, Set
from lark import Lark, Transformer
from functools import reduce
from operator import add, mul
from fractions import Fraction


class Var:
    name: str
    _is_bool: bool
    _types_dict: Optional[Union[bool, Dict[str, bool]]]

    def __init__(self, name: str, data: Union[bool, Dict[str, bool]]):
        self.name = name
        if isinstance(data, dict):
            self._types_dict = data
            self._frozen = False
            self._is_bool = False
        else:
            self._types_dict = None
            self._frozen = True
            self._is_bool = data

    @property
    def is_bool(self) -> bool:
        if self._frozen:
            return self._is_bool
        else:
            assert isinstance(self._types_dict, dict)
            return self._types_dict[self.name]

    def __repr__(self) -> str:
        return self.name

    def __lt__(self, other: 'Var') -> bool:
        if int(self.is_bool) == int(other.is_bool):
            return self.name < other.name
        return int(self.is_bool) < int(other.is_bool)

    def __hash__(self) -> int:
        # Variables with the same name map to the same type
        # within the same graph, so no need to include is_bool
        # in the hash.
        return int(hash(self.name))

    def __eq__(self, other: object) -> bool:
        return self.__hash__() == other.__hash__()

    def freeze(self) -> None:
        if not self._frozen:
            assert isinstance(self._types_dict, dict)
            self._is_bool = self._types_dict[self.name]
            self._frozen = True
            self._types_dict = None

    def __copy__(self) -> 'Var':
        if self._frozen:
            return Var(self.name, self.is_bool)
        else:
            assert isinstance(self._types_dict, dict)
            return Var(self.name, self._types_dict)

    def __deepcopy__(self, _memo: object) -> 'Var':
        return self.__copy__()

class Term:
    vars: List[Tuple[Var, int]]

    def __init__(self, vars: List[Tuple[Var,int]]) -> None:
        self.vars = vars

    def freeze(self) -> None:
        for var, _ in self.vars:
            var.freeze()

    def free_vars(self) -> Set[Var]:
        return set(var for var, _ in self.vars)

    def __repr__(self) -> str:
        vs = []
        for v, c in self.vars:
            if c == 1:
                vs.append(f'{v}')
            else:
                vs.append(f'{v}^{c}')
        return '*'.join(vs)

    def __mul__(self, other: 'Term') -> 'Term':
        vs = dict()
        for v, c in self.vars + other.vars:
            if v not in vs: vs[v] = c
            else: vs[v] += c
            # TODO deal with fractional / symbolic powers
            if v.is_bool and c > 1:
                vs[v] = 1
        return Term([(v, c) for v, c in vs.items()])

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.vars)))

    def __eq__(self, other: object) -> bool:
        return self.__hash__() == other.__hash__()

    def __lt__(self, other: 'Term') -> bool:
        if len(self.vars) != len(other.vars):
            return len(self.vars) < len(other.vars)
        for (v1, c1), (v2, c2) in zip(sorted(self.vars), sorted(other.vars)):
            if v1 != v2: return v1 < v2
            if c1 != c2: return c1 < c2
        return False

    def substitute(self, var_map: Dict[Var, Union[float, complex, 'Fraction']]) -> Tuple[Union[float, complex, 'Fraction'], 'Term']:
        """Substitute variables in the term with the given values. Returns a tuple
        of the coefficient and the new term.
        """
        coeff: Union[float, complex, 'Fraction'] = 1.
        new_vars = []
        for v, c in self.vars:
            if v in var_map:
                coeff *= var_map[v] ** c
            else:
                new_vars.append((v, c))
        return (coeff, Term(new_vars))


class Poly:
    terms: List[Tuple[Union[int, float, complex, Fraction], Term]]

    def __init__(self, terms: List[Tuple[Union[int, float, complex, Fraction], Term]]) -> None:
        self.terms = terms

    def freeze(self) -> None:
        for _, term in self.terms:
            term.freeze()

    def free_vars(self) -> Set[Var]:
        output = set()
        for _, term in self.terms:
            output.update(term.free_vars())
        return output

    def __add__(self, other: Union['Poly', Fraction, int, float, complex]) -> 'Poly':
        if isinstance(other, (int, float, complex, Fraction)):
            other = Poly([(other, Term([]))])
        counter = dict()
        for c, t in self.terms + other.terms:
            if t not in counter: counter[t] = c
            else: counter[t] += c
            counter_t = counter[t] # need to assign to variable to avoid type errors
            if not isinstance(counter_t, complex) and len(t.vars) > 0 and all(tt[0].is_bool for tt in t.vars):
                counter[t] = counter_t % 2

        # remove terms with coefficient 0
        for t in list(counter.keys()):
            if counter[t] == 0:
                del counter[t]
        return Poly([(c, t) for t, c in counter.items()])

    __radd__ = __add__

    def __neg__(self) -> 'Poly':
        return Poly([(-c, t) for c, t in self.terms])

    def __sub__(self, other: Union['Poly', Fraction, int, float, complex]) -> 'Poly':
        return self + (-other)

    def __rsub__(self, other: Union['Poly', Fraction, int, float, complex]) -> 'Poly':
        return other + (-self)

    def __mul__(self, other: Union['Poly', Fraction, int, float, complex]) -> 'Poly':
        if isinstance(other, (int, float, complex, Fraction)):
            other = Poly([(other, Term([]))])
        p = Poly([])
        for c1, t1 in self.terms:
            for c2, t2 in other.terms:
                p += Poly([(c1 * c2, t1 * t2)])
        return p

    __rmul__ = __mul__

    def __truediv__(self, other: Union['Poly', Fraction, int, float, complex]) -> 'Poly':
        if isinstance(other, (int, float, complex, Fraction)):
            other = Poly([(other, Term([]))])
        if len(other.terms) == 0:
            raise ZeroDivisionError("division by zero")
        quotient = Poly([])
        while len(self.terms) > 0 and self.degree >= other.degree:
            leading_term_dividend = sorted(self.terms)[0][1]
            leading_term_divisor = sorted(other.terms)[0][1]
            coeff = sorted(self.terms)[0][0] / sorted(other.terms)[0][0]
            new_term_quotient_vars = [(var, exp - dict(leading_term_divisor.vars).get(var, 0)) for var, exp in leading_term_dividend.vars]
            new_term_quotient = (coeff, Term(new_term_quotient_vars))
            quotient.terms.append(new_term_quotient)
            self -= other * Poly([new_term_quotient])
        return quotient

    def __pow__(self, other: int) -> 'Poly':
        if other < 0:
            return Poly([(1, Term([]))]) / (self ** (-other))
        if other == 0:
            return Poly([(1, Term([]))])
        if other == 1:
            return self
        return self * (self ** (other - 1))

    def __mod__(self, other: int) -> 'Poly':
        return Poly([(c % other, t) for c, t in self.terms if not isinstance(c, complex)])

    def __repr__(self) -> str:
        return f'Poly({str(self)})'

    def __str__(self) -> str:
        ts = []
        for c, t in self.terms:
            if t == Term([]):
                ts.append(f'{c}')
            elif c == 1:
                ts.append(f'{t}')
            else:
                ts.append(f'{c}{t}')
        return ' + '.join(ts)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (int, float, Fraction)):
            if other == 0:
                other = Poly([])
            else:
                other = Poly([(other, Term([]))])
        if not isinstance(other, Poly): return False
        return set(self.terms) == set(other.terms)

    def __lt__(self, other: Union['Poly', Fraction, int, float, complex]) -> bool:
        if isinstance(other, (int, float, complex, Fraction)):
            other = Poly([(other, Term([]))])
        if len(self.terms) != len(other.terms):
            return len(self.terms) < len(other.terms)
        for (c1, t1), (c2, t2) in zip(sorted(self.terms), sorted(other.terms)):
            if t1 != t2: return t1 < t2
            if c1 != c2 and not isinstance(c1, complex) and not isinstance(c2, complex):
                return c1 < c2
        return False

    def __le__(self, other: Union['Poly', Fraction, int, float, complex]) -> bool:
        return self == other or self < other

    def __gt__(self, other: Union['Poly', Fraction, int, float, complex]) -> bool:
        return not self <= other

    def __ge__(self, other: Union['Poly', Fraction, int, float, complex]) -> bool:
        return not self < other

    def __hash__(self) -> int:
        return hash(tuple(sorted(self.terms)))

    @property
    def degree(self) -> int:
        powers = [sum(c for _, c in t.vars) for coeff, t in self.terms if coeff != 0]
        if powers:
            return max(powers)
        return 0

    @property
    def is_pauli(self) -> bool:
        for c, t in self.terms:
            if isinstance(c, complex):
                return False
            if not all(v.is_bool for v, _ in t.vars):
                return False
            if c % 1 != 0:
                return False
        return True

    @property
    def is_clifford(self) -> bool:
        for c, t in self.terms:
            if isinstance(c, complex):
                return False
            if not all(v.is_bool for v, _ in t.vars):
                return False
            if t.vars: # Contains a Boolean variable
               if c % 1 != 0: # With weight not equal to 1
                return False
            else:
                if c*2 % 1 != 0: # Variable-free term with weight not equal to 1/2
                    return False
        return True

    @property
    def numerator(self) -> int:
        raise NotImplementedError('numerator not implemented for symbolic Poly')

    @property
    def denominator(self) -> int:
        raise NotImplementedError('denominator not implemented for symbolic Poly')

    def copy(self) -> 'Poly':
        return Poly([(c, t) for c, t in self.terms])

    def substitute(self, var_map: Dict[Var, Union[float, complex, 'Fraction']]) -> 'Poly':
        """Substitute variables in the polynomial with the given values."""
        p = Poly([])
        for c, t in self.terms:
            coeff, term = t.substitute(var_map)
            p += Poly([(c * coeff, term)])
        return p

def new_var(name: str, types_dict: Union[bool, Dict[str, bool]]) -> Poly:
    return Poly([(1, Term([(Var(name, types_dict), 1)]))])

def new_const(coeff: Union[int, Fraction]) -> Poly:
    return Poly([(coeff, Term([]))])


poly_grammar = Lark("""
    start      : "(" start ")" | term ("+" term)*
    term       : (intf | frac)? factor ("*" factor)*
    ?factor    : intf | frac | pi | pifrac | var
    var        : CNAME
    intf       : INT
    pi         : "\\pi" | "pi"
    frac       : INT "/" INT
    pifrac     : [INT] pi "/" INT

    %import common.INT
    %import common.CNAME
    %import common.WS
    %ignore WS
    """,
    parser='lalr',
    maybe_placeholders=True)

class PolyTransformer(Transformer):
    def __init__(self, new_var: Callable[[str], Poly]):
        super().__init__()

        self._new_var = new_var

    def start(self, items: List[Poly]) -> Poly:
        return reduce(add, items)

    def term(self, items: List[Poly]) -> Poly:
        return reduce(mul, items)

    def var(self, items: List[Any]) -> Poly:
        v = str(items[0])
        return self._new_var(v)

    def pi(self, _: List[Any]) -> Poly:
        return new_const(1)

    def intf(self, items: List[Any]) -> Poly:
        return new_const(int(items[0]))

    def frac(self, items: List[Any]) -> Poly:
        return new_const(Fraction(int(items[0]), int(items[1])))

    def pifrac(self, items: List[Any]) -> Poly:
        numerator = int(items[0]) if items[0] else 1
        return new_const(Fraction(numerator, int(items[2])))

def parse(expr: str, new_var: Callable[[str], Poly]) -> Poly:
    tree = poly_grammar.parse(expr)
    return PolyTransformer(new_var).transform(tree)



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module implements :class:`Architecture` aware routing methods for :class:`Circuit` s.
"""

__all__ = [
    # Architecture objects
    "architectures",
    "create_architecture",
    "Architecture",

    # Qubit parity map trackers
    "Parity",
    "CNOT_tracker",

    # Cnot mapping
    "ElimMode",
    "CostMetric",
    "FitnessFunction",
    "gauss",
    "permuted_gauss",
    "sequential_gauss",
    "steiner_gauss",
    "rec_steiner_gauss",

    # Phase polynomial routing
    "RoutingMethod",
    "RootHeuristic",
    "SplitHeuristic",
    "route_phase_poly",
]

from .architecture import Architecture, architectures, create_architecture
from .cnot_mapper import ElimMode, CostMetric, FitnessFunction, gauss, permuted_gauss, sequential_gauss
from .parity_maps import Parity, CNOT_tracker
from .steiner import steiner_gauss, rec_steiner_gauss
from .phase_poly import RoutingMethod, RootHeuristic, SplitHeuristic, route_phase_poly


# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from typing import Any
import numpy as np
import importlib

# try:
#     import quimb as qu # type:ignore
#     import quimb.tensor as qtn #type:ignore
# except ImportError:

# load quimb modules on first call to to_quimb_tensor
qu: Any = None
qtn: Any = None

from .utils import EdgeType, VertexType
from .graph.base import BaseGraph
from .simplify import to_gh
from .symbolic import Poly


def to_quimb_tensor(g: BaseGraph) -> "qtn.TensorNetwork": # type:ignore
    """Converts tensor network representing the given :func:`pyzx.graph.Graph`.
    Pretty printing: to_tensor(g).draw(color = ['V', 'H'])
    
    Args:
        g: graph to be converted."""

    global qu, qtn
    if qu is None:
        qu = importlib.import_module('quimb')
        qtn = importlib.import_module('quimb.tensor')

    # copying a graph guarantees consecutive indices, which are needed for the tensor net
    g = g.copy()

    # only Z spiders are handled below
    to_gh(g)
    
    tensors = []

    # Here we have phase tensors corresponding to Z-spiders with only one output and no input.
    for v in g.vertices():
        if g.type(v) == VertexType.Z and g.phase(v) != 0:
            phase = g.phase(v)
            if isinstance(phase, Poly):
                raise NotImplementedError("Quimb does not support symbolic phases")
            tensors.append(qtn.Tensor(data = [1, np.exp(1j * np.pi * phase)],
                                      inds = (f'{v}',),
                                      tags = ("V",)))
    

    # Hadamard or Kronecker tensors, one for each edge of the diagram.
    for i, edge in enumerate(g.edges()):
        x, y = edge
        isHadamard = g.edge_type(edge) == EdgeType.HADAMARD
        t = qtn.Tensor(data = qu.hadamard() if isHadamard else np.array([1, 0, 0, 1]).reshape(2, 2),
                       inds = (f'{x}', f'{y}'),
                       tags = ("H",) if isHadamard else ("N",))
        tensors.append(t)

    # TODO: This is not taking care of all the stuff that can be in g.scalar
    # In particular, it doesn't check g.scalar.phasenodes
    # TODO: This will give the wrong tensor when g.scalar.is_zero == True.
    # Grab the float factor and exponent from the scalar
    phase = g.scalar.phase
    if isinstance(phase, Poly):
        raise NotImplementedError("Quimb does not support symbolic phases")
    scalar_float = np.exp(1j * np.pi * phase) * g.scalar.floatfactor.to_complex()
    for node in g.scalar.phasenodes:    # Each node is a Fraction
        if isinstance(node, Poly):
            raise NotImplementedError("Quimb does not support symbolic phases")
        scalar_float *= 1 + np.exp(1j * np.pi * node)
    scalar_exp = math.log10(math.sqrt(2)) * g.scalar.power2

    # If the TN is empty, create a single 0-tensor with scalar factor, otherwise
    # multiply the scalar into one of the tensors.
    if len(tensors) == 0:
        tensors.append(qtn.Tensor(data = scalar_float,
                                  inds = (),
                                  tags = ("S",)))
    else:
        tensors[0].modify(data = tensors[0].data * scalar_float)


    network = qtn.TensorNetwork(tensors)

    # the exponent can be very large, so distribute it evenly through the TN
    network.exponent = scalar_exp
    network.distribute_exponent()
    return network



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
NEW VERSION

This module contains several rules more easily applied interactively to ZX
diagrams. The emphasis is more on ease of use and simplicity than performance.

Rules are given as functions that take as input a vertex or a pair of vertices
to fix the location the rule is applied. They then apply the rule and return
True if the rule indeed applies at this location, otherwise they return false.

Most rules also have a companion function check_RULENAME, which only checks
whether the rule applies at the given location and doesn't actually apply
the rule.
"""

__all__ = ['color_change_diagram',
        'check_color_change',
        'color_change',
        'check_copy_X',
        'copy_X',
        'check_copy_Z',
        'copy_Z',
        'check_pi_commute_X',
        'pi_commute_X',
        'check_pi_commute_Z',
        'pi_commute_Z',
        'check_strong_comp',
        'strong_comp',
        'check_fuse',
        'fuse',
        'check_remove_id',
        'remove_id']

from typing import Tuple, List
from .editor_actions import bialgebra
from pyzx_param.graph.base import BaseGraph, VT, ET
from .rules import apply_rule, w_fusion, z_to_z_box
from pyzx_param.utils import (EdgeType, VertexType, get_w_io, get_z_box_label, is_pauli,
                    set_z_box_label, vertex_is_w, vertex_is_z_like, toggle_vertex, toggle_edge)

def color_change_diagram(g: BaseGraph[VT,ET]):
    """Color-change an entire diagram by applying Hadamards to the inputs and ouputs."""
    for v in g.vertices():
        if g.type(v) == VertexType.BOUNDARY:
            if g.vertex_degree(v) != 1:
                raise ValueError("Boundary should only have 1 neighbor.")
            for e in g.incident_edges(v):
                g.set_edge_type(e, toggle_edge(g.edge_type(e)))
        elif check_color_change(g, v):
            g.set_type(v, toggle_vertex(g.type(v)))

def check_color_change(g: BaseGraph[VT,ET], v: VT) -> bool:
    return (g.type(v) == VertexType.Z or g.type(v) == VertexType.X)

def color_change(g: BaseGraph[VT,ET], v: VT) -> bool:
    """Color-change a vertex by applying Hadamards to all incident edges. Must be either a Z- or X- vertex"""
    if not (g.type(v) == VertexType.Z or g.type(v) == VertexType.X):
        return False

    g.set_type(v, toggle_vertex(g.type(v)))
    for e in g.incident_edges(v):
        g.set_edge_type(e, toggle_edge(g.edge_type(e)))

    return True

def check_strong_comp(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if not (((g.type(v1) == VertexType.X and g.type(v2) == VertexType.Z) or
             (g.type(v1) == VertexType.Z and g.type(v2) == VertexType.X)) and
            is_pauli(g.phase(v1)) and
            is_pauli(g.phase(v2)) and
            g.connected(v1,v2) and
            EdgeType.SIMPLE in [g.edge_type(edge) for edge in g.edges(v1,v2)]):
        return False
    return True

def strong_comp(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if not check_strong_comp(g, v1, v2): return False

    etab, rem_verts, rem_edges, check_isolated_vertices = bialgebra(g, [(v1, v2)])
    g.remove_edges(rem_edges)
    g.remove_vertices(rem_verts)
    g.add_edge_table(etab)
    return True

def check_copy_X(g: BaseGraph[VT,ET], v: VT) -> bool:
    if not (g.vertex_degree(v) == 1 and
            g.type(v) == VertexType.X and
            is_pauli(g.phase(v))):
        return False
    nv = next(iter(g.neighbors(v)))
    if not (g.type(nv) == VertexType.Z and
            g.edge_type(g.edge(v,nv)) == EdgeType.SIMPLE):
        return False
    return True

def copy_X(g: BaseGraph[VT,ET], v: VT) -> bool:
    if not check_copy_X(g, v): return False
    nv = next(iter(g.neighbors(v)))
    strong_comp(g, v, nv)

    return True

def check_pi_commute_Z(g: BaseGraph[VT, ET], v: VT) -> bool:
    return g.type(v) == VertexType.Z

def pi_commute_Z(g: BaseGraph[VT, ET], v: VT) -> bool:
    if not check_pi_commute_Z(g, v): return False
    g.set_phase(v, -g.phase(v))
    ns = g.neighbors(v)
    for w in ns:
        e = g.edge(v, w)
        et = g.edge_type(e)
        if ((g.type(w) == VertexType.Z and et == EdgeType.HADAMARD) or
            (g.type(w) == VertexType.X and et == EdgeType.SIMPLE)):
            g.add_to_phase(w, 1)
        else:
            g.remove_edge(e)
            c = g.add_vertex(VertexType.X,
                    qubit=0.5*(g.qubit(v) + g.qubit(w)),
                    row=0.5*(g.row(v) + g.row(w)))
            g.add_edge((v, c))
            g.add_edge((c, w), edgetype=et)
    return True

def check_pi_commute_X(g: BaseGraph[VT,ET], v: VT) -> bool:
    color_change_diagram(g)
    b = check_pi_commute_Z(g, v)
    color_change_diagram(g)
    return b

def pi_commute_X(g: BaseGraph[VT,ET], v: VT) -> bool:
    color_change_diagram(g)
    b = pi_commute_Z(g, v)
    color_change_diagram(g)
    return b

def check_copy_Z(g: BaseGraph[VT,ET], v: VT) -> bool:
    color_change_diagram(g)
    b = check_copy_X(g, v)
    color_change_diagram(g)
    return b

def copy_Z(g: BaseGraph, v: VT) -> bool:
    color_change_diagram(g)
    b = copy_X(g, v)
    color_change_diagram(g)
    return b

def check_fuse(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if check_fuse_w(g, v1, v2):
        return True
    if not (g.connected(v1,v2) and
            ((g.type(v1) == VertexType.X and g.type(v2) == VertexType.X) or
             (vertex_is_z_like(g.type(v1)) and vertex_is_z_like(g.type(v2)))) and
            EdgeType.SIMPLE in [g.edge_type(edge) for edge in g.edges(v1,v2)]):
        return False
    return True

def fuse(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if not check_fuse(g, v1, v2): return False
    if vertex_is_w(g.type(v1)):
        return fuse_w(g, v1, v2)
    if g.type(v1) == VertexType.Z_BOX or g.type(v2) == VertexType.Z_BOX:
        if g.type(v1) == VertexType.Z:
            z_to_z_box(g, [v1])
        if g.type(v2) == VertexType.Z:
            z_to_z_box(g, [v2])
        set_z_box_label(g, v1, get_z_box_label(g, v1) * get_z_box_label(g, v2))
    else:
        g.add_to_phase(v1, g.phase(v2), g.get_params(v2))
    for e in g.incident_edges(v2):
        source, target = g.edge_st(e)
        other_vertex = source if source != v2 else target
        if source != v2:
            other_vertex = source
        elif target != v2:
            other_vertex = target
        else: #self-loop
            other_vertex = v1
        if other_vertex == v1 and g.edge_type(e) == EdgeType.SIMPLE:
            continue
        g.add_edge((v1,other_vertex), edgetype=g.edge_type(e))
    g.remove_vertex(v2)
    return True

def check_fuse_w(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if vertex_is_w(g.type(v1)) and vertex_is_w(g.type(v2)):
        v1_in, v1_out = get_w_io(g, v1)
        v2_in, v2_out = get_w_io(g, v2)
        if (g.connected(v1_in, v2_out) and g.edge_type(g.edge(v1_in, v2_out)) == EdgeType.SIMPLE) or \
           (g.connected(v2_in, v1_out) and g.edge_type(g.edge(v2_in, v1_out)) == EdgeType.SIMPLE):
            return True
    return False

def fuse_w(g: BaseGraph[VT,ET], v1: VT, v2: VT) -> bool:
    if not check_fuse_w(g, v1, v2): return False
    v1_in, v1_out = get_w_io(g, v1)
    v2_in, v2_out = get_w_io(g, v2)
    if not g.connected(v1_out, v2_in):
        g.set_position(v2_in, g.qubit(v1_in), g.row(v1_in))
        g.set_position(v2_out, g.qubit(v1_out), g.row(v1_out))
    apply_rule(g, w_fusion, [(v2, v1)])
    return True

def check_remove_id(g: BaseGraph[VT,ET], v: VT) -> bool:
    if not g.vertex_degree(v) == 2:
        return False
    if g.type(v) == VertexType.Z_BOX and get_z_box_label(g, v) == 1:
        return True
    elif g.type(v) != VertexType.Z_BOX and g.phase(v) == 0:
        return True
    return False

def remove_id(g: BaseGraph[VT,ET], v: VT) -> bool:
    if vertex_is_w(g.type(v)):
        return remove_id_w(g, v)
    if not check_remove_id(g, v):
        return False

    neighbors = list(g.neighbors(v))
    if len(neighbors) == 2:
        v1, v2 = neighbors[0], neighbors[1]
    else: # self loop
        v1, v2 = neighbors[0], neighbors[0]
    g.add_edge((v1,v2), edgetype=EdgeType.SIMPLE
            if g.edge_type(g.edge(v,v1)) == g.edge_type(g.edge(v,v2))
            else EdgeType.HADAMARD)
    g.remove_vertex(v)

    return True

def check_remove_id_w(g: BaseGraph[VT,ET], v: VT) -> bool:
    w_in, w_out = get_w_io(g, v)
    if g.vertex_degree(w_out) == 2:
        return True
    return False

def remove_id_w(g: BaseGraph[VT,ET], v: VT) -> bool:
    if not check_remove_id_w(g, v):
        return False
    w_in, w_out = get_w_io(g, v)
    v1 = [n for n in g.neighbors(w_out) if n != w_in][0]
    v2 = [n for n in g.neighbors(w_in) if n != w_out][0]
    g.add_edge((v1,v2), edgetype=EdgeType.SIMPLE
        if g.edge_type(g.edge(w_out, v1)) == g.edge_type(g.edge(w_in, v2))
        else EdgeType.HADAMARD)
    g.remove_vertex(w_in)
    g.remove_vertex(w_out)
    return True



# PyZX - Python library for quantum circuit rewriting
#       and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections import Counter
from fractions import Fraction
from typing import Tuple, Dict, Set, Union, Any

from .base import BaseGraph

from ..utils import VertexType, EdgeType, FractionLike, FloatInt, vertex_is_zx_like, vertex_is_z_like, set_z_box_label, get_z_box_label

class Edge:
    """A structure for storing the number of simple and number of Hadamard edges
    between two vertices"""
    s: int
    h: int
    w_io: int

    def __init__(self, s: int=0, h: int=0, w_io: int=0):
        self.s = s
        self.h = h
        self.w_io = w_io

    def add(self, s: int=0, h: int=0, w_io: int=0):
        self.s += s
        self.h += h
        self.w_io += w_io
        if self.s < 0 or self.h < 0:
            raise ValueError('Cannot have negative edges')
        if self.w_io not in (0,1):
            raise ValueError('Invalid number of W-IO edges')
        if self.w_io == 1 and self.s + self.h > 0:
            raise ValueError('Cannot have W-IO edge and other edges')

    def remove(self, s: int=0, h: int=0, w_io: int=0):
        self.add(s=-s, h=-h, w_io=-w_io)

    def is_empty(self) -> bool:
        return self.s == 0 and self.h == 0 and self.w_io == 0

    def get_edge_count(self, ty: EdgeType) -> int:
        if ty == EdgeType.SIMPLE: return self.s
        elif ty == EdgeType.HADAMARD: return self.h
        else: return self.w_io

class Multigraph(BaseGraph[int,Tuple[int,int,EdgeType]]):
    """Purely Pythonic multigraph implementation of :class:`~graph.base.BaseGraph`."""
    backend = 'multigraph'

    #The documentation of what these methods do
    #can be found in base.BaseGraph
    def __init__(self) -> None:
        BaseGraph.__init__(self)
        self.graph: Dict[int,Dict[int,Edge]]   = dict()
        self._auto_simplify: bool                       = True
        self._vindex: int                               = 0
        self.nedges: int                                = 0
        self.ty: Dict[int,VertexType]                   = dict()
        self._phase: Dict[int, FractionLike]            = dict()
        self._qindex: Dict[int, FloatInt]               = dict()
        self._maxq: FloatInt                            = -1
        self._rindex: Dict[int, FloatInt]               = dict()
        self._maxr: FloatInt                            = -1
        self._grounds: Set[int]                         = set()

        self._vdata: Dict[int,Any]                      = dict()
        self._edata: Dict[Tuple[int,int,EdgeType], Dict[str, Any]] = dict()
        self._inputs: Tuple[int, ...]                   = tuple()
        self._outputs: Tuple[int, ...]                  = tuple()

    def clone(self) -> 'Multigraph':
        cpy = Multigraph()
        for v, d in self.graph.items():
            cpy.graph[v] = d.copy()
        cpy._vindex = self._vindex
        cpy.nedges = self.nedges
        cpy.ty = self.ty.copy()
        cpy._phase = self._phase.copy()
        cpy._qindex = self._qindex.copy()
        cpy._maxq = self._maxq
        cpy._rindex = self._rindex.copy()
        cpy._maxr = self._maxr
        cpy._vdata = self._vdata.copy()
        cpy._edata = {k: v.copy() for k, v in self._edata.items()}
        cpy.scalar = self.scalar.copy()
        cpy._inputs = tuple(list(self._inputs))
        cpy._outputs = tuple(list(self._outputs))
        cpy.track_phases = self.track_phases
        cpy.phase_index = self.phase_index.copy()
        cpy.phase_master = self.phase_master
        cpy.phase_mult = self.phase_mult.copy()
        cpy.max_phase_index = self.max_phase_index
        return cpy

    def set_auto_simplify(self, s: bool):
        """Automatically remove parallel edges as edges are added"""
        self._auto_simplify = s

    def get_auto_simplify(self):
        return self._auto_simplify

    def multigraph(self):
        return False

    def vindex(self): return self._vindex
    def depth(self):
        if self._rindex: self._maxr = max(self._rindex.values())
        else: self._maxr = -1
        return self._maxr
    def qubit_count(self):
        if self._qindex: self._maxq = max(self._qindex.values())
        else: self._maxq = -1
        return self._maxq + 1

    def inputs(self):
        return self._inputs

    def num_inputs(self):
        return len(self._inputs)

    def set_inputs(self, inputs):
        self._inputs = inputs

    def outputs(self):
        return self._outputs

    def num_outputs(self):
        return len(self._outputs)

    def set_outputs(self, outputs):
        self._outputs = outputs

    def add_vertices(self, amount):
        for i in range(self._vindex, self._vindex + amount):
            self.graph[i] = dict()
            self.ty[i] = VertexType.BOUNDARY
            self._phase[i] = 0
        self._vindex += amount
        return range(self._vindex - amount, self._vindex)
    def add_vertex_indexed(self, v):
        """Adds a vertex that is guaranteed to have the chosen index (i.e. 'name').
        If the index isn't available, raises a ValueError.
        This method is used in the editor to support undo, which requires vertices
        to preserve their index."""
        if v in self.graph: raise ValueError("Vertex with this index already exists")
        if v >= self._vindex: self._vindex = v+1
        self.graph[v] = dict()
        self.ty[v] = VertexType.BOUNDARY
        self._phase[v] = 0

    def add_edges(self, edge_pairs, edgetype=EdgeType.SIMPLE):
        for ep in edge_pairs: self.add_edge(ep, edgetype)

    def add_edge(self, edge_pair, edgetype=EdgeType.SIMPLE):
        self.nedges += 1
        s,t = edge_pair
        if not t in self.graph[s]:
            e = Edge()
            self.graph[s][t] = e
            self.graph[t][s] = e
        else:
            e = self.graph[s][t]

        if edgetype == EdgeType.SIMPLE: e.add(s=1)
        elif edgetype == EdgeType.HADAMARD: e.add(h=1)
        else: e.add(w_io=1)

        if self._auto_simplify: # This currently can keep a parallel regular and Hadamard edge
            t1 = self.ty[s]
            t2 = self.ty[t]
            if (vertex_is_zx_like(t1) and vertex_is_zx_like(t2)):
                if s == t: # turn self-loops in pi phases
                    e.s = 0
                    if e.h % 2 == 1:
                        if t1 == VertexType.Z_BOX:
                            set_z_box_label(self, s, get_z_box_label(self, s) * -1)
                        else:
                            self.add_to_phase(s, 1)
                    self.scalar.add_power(-e.h)
                    self.nedges = self.nedges - e.h
                    e.h = 0
                else: # apply spider and hopf to merge/cancel parallel edges
                    if t1 == t2:
                        if e.s > 0:
                            self.nedges = self.nedges - e.s + 1
                            e.s = 1
                        if e.h > 0:
                            self.nedges = self.nedges - (e.h - e.h % 2)
                            self.scalar.add_power(-2 * (e.h - (e.h % 2)))
                            e.h = e.h % 2
                    else:
                        if e.h > 0:
                            self.nedges = self.nedges - e.h + 1
                            e.h = 1
                        if e.s > 0:
                            self.nedges = self.nedges - (e.s - e.s % 2)
                            self.scalar.add_power(-2 * (e.s - (e.s % 2)))
                            e.s = e.s % 2
            if e.is_empty():
                del self.graph[s][t]
                if s != t:
                    del self.graph[t][s]
                self._edata.pop((s, t, edgetype), None)

        return (s,t, edgetype) if s <= t else (t,s,edgetype)

    def remove_vertices(self, vertices):
        for v in vertices:
            vs = list(self.graph[v])
            # remove all edges
            for v1 in vs:
                e = self.graph[v][v1]
                self.nedges -= e.s + e.h
                # Remove all edata for all edge types between v and v1
                for ty in EdgeType:
                    self._edata.pop((min(v, v1), max(v, v1), ty), None)
                del self.graph[v][v1]
                if v != v1: del self.graph[v1][v]
            # remove the vertex
            del self.graph[v]
            del self.ty[v]
            del self._phase[v]
            if v in self._inputs:
                self._inputs = tuple(u for u in self._inputs if u != v)
            if v in self._outputs:
                self._outputs = tuple(u for u in self._outputs if u != v)
            try: del self._qindex[v]
            except: pass
            try: del self._rindex[v]
            except: pass
            try: del self.phase_index[v]
            except: pass
            self._grounds.discard(v)
            self._vdata.pop(v,None)
        self._vindex = max(self.vertices(),default=0) + 1

    def remove_vertex(self, vertex):
        self.remove_vertices([vertex])

    def remove_edges(self, edges):
        for e in edges:
            self.remove_edge(e)

    def remove_edge(self, edge):
        s,t,ty = edge
        e = self.graph[s][t]
        if ty == EdgeType.SIMPLE: e.remove(s=1)
        elif ty == EdgeType.HADAMARD: e.remove(h=1)
        else: e.remove(w_io=1)

        if e.is_empty():
            del self.graph[s][t]
            if s != t: del self.graph[t][s]
            self._edata.pop(edge, None)

        self.nedges -= 1

    def num_vertices(self):
        return len(self.graph)

    def num_edges(self, s=None, t=None, et=None):
        if s != None or t != None:
            if et == None:
                return len(list(self.edges(s, t)))
            if not t in self.graph[s]:
                return 0
            if et == EdgeType.SIMPLE:
                return self.graph[s][t].s
            elif et == EdgeType.HADAMARD:
                return self.graph[s][t].h
            elif et == EdgeType.W_IO:
                return self.graph[s][t].w_io
            else:
                raise ValueError("Unkown EdgeType: %s" % repr(et))
        else:
            return self.nedges

    def vertices(self):
        return self.graph.keys()

    def vertices_in_range(self, start, end):
        """Returns all vertices with index between start and end
        that only have neighbours whose indices are between start and end"""
        for v in self.graph.keys():
            if not start<v<end: continue
            if all(start<v2<end for v2 in self.graph[v]):
                yield v

    def edges(self, s=None, t=None):
        if s == None:
            for v0,adj in self.graph.items():
                for v1, e in adj.items():
                    if v1 >= v0:
                        for _ in range(e.s): yield (v0, v1, EdgeType.SIMPLE)
                        for _ in range(e.h): yield (v0, v1, EdgeType.HADAMARD)
                        for _ in range(e.w_io): yield (v0, v1, EdgeType.W_IO)
        elif t != None:
            s, t = (s, t) if s < t else (t, s)
            if t not in self.graph[s]:
                return
            e = self.graph[s][t]
            for _ in range(e.s): yield (s, t, EdgeType.SIMPLE)
            for _ in range(e.h): yield (s, t, EdgeType.HADAMARD)
            for _ in range(e.w_io): yield (s, t, EdgeType.W_IO)

    # def edges_in_range(self, start, end, safe=False):
    #    """like self.edges, but only returns edges that belong to vertices
    #    that are only directly connected to other vertices with
    #    index between start and end.
    #    If safe=True then it also checks that every neighbour is only connected to vertices with the right index"""
    #    if not safe:
    #        for v0,adj in self.graph.items():
    #            if not (start<v0<end): continue
    #            #verify that all neighbours are in range
    #            if all(start<v1<end for v1 in adj):
    #                for v1 in adj:
    #                    if v1 > v0: yield (v0,v1)
    #    else:
    #        for v0,adj in self.graph.items():
    #            if not (start<v0<end): continue
    #            #verify that all neighbours are in range, and that each neighbour
    #            # is only connected to vertices that are also in range
    #            if all(start<v1<end for v1 in adj) and all(all(start<v2<end for v2 in self.graph[v1]) for v1 in adj):
    #                for v1 in adj:
    #                    if v1 > v0:
    #                        yield (v0,v1)

    def edge(self, s, t, et: EdgeType=EdgeType.SIMPLE):
        s, t = (s, t) if s < t else (t, s)
        e = self.graph[s][t]
        if e.get_edge_count(et):
            return (s, t, et)
        return next(self.edges(s, t))


    def edge_set(self):
        return Counter(self.edges())

    def edge_st(self, edge):
        return (edge[0], edge[1])

    def neighbors(self, vertex):
        return self.graph[vertex].keys()

    def vertex_degree(self, vertex):
        d = 0
        for e in self.graph[vertex].values():
            for ty in EdgeType:
                d += e.get_edge_count(ty)
        return d

    def incident_edges(self, vertex):
        return list(itertools.chain.from_iterable(
            self.edges(vertex, v1) for v1 in self.graph[vertex]
        ))

    def connected(self,v1,v2):
        return v2 in self.graph[v1]

    def edge_type(self, e):
        if len(e) == 2:
            edges = list(self.edges(e[0],e[1]))
            if len(edges) == 0:
                raise ValueError('Edge not found')
            if len(edges) > 1:
                # if all edges are of the same type, return that type
                if all(e[2] == edges[0][2] for e in edges):
                    return edges[0][2]
                else:
                    raise ValueError('Cannot determine edge type')
            e = edges[0]
        return e[2]

    def set_edge_type(self, edge, t):
        v1,v2,ty = edge
        if ty != t:
            e = self.graph[v1][v2]
            # decrement the old type and increment the new type
            if ty == EdgeType.SIMPLE: e.add(s=-1)
            elif ty == EdgeType.HADAMARD: e.add(h=-1)
            else: e.add(w_io=-1)
            if t == EdgeType.SIMPLE: e.add(s=1)
            elif t == EdgeType.HADAMARD: e.add(h=1)
            else: e.add(w_io=1)

    def type(self, vertex):
        return self.ty[vertex]
    def types(self):
        return self.ty
    def set_type(self, vertex, t):
        self.ty[vertex] = t

    def phase(self, vertex):
        return self._phase.get(vertex,Fraction(1))
    def phases(self):
        return self._phase
    def set_phase(self, vertex, phase):
        try:
            self._phase[vertex] = Fraction(phase) % 2
        except Exception:
            self._phase[vertex] = phase
    def add_to_phase(self, vertex, phase):
        old_phase = self._phase.get(vertex, Fraction(1))
        try:
            self._phase[vertex] = (old_phase + Fraction(phase)) % 2
        except Exception:
            self._phase[vertex] = old_phase + phase
    def qubit(self, vertex):
        return self._qindex.get(vertex,-1)
    def qubits(self):
        return self._qindex
    def set_qubit(self, vertex, q):
        if q > self._maxq: self._maxq = q
        self._qindex[vertex] = q

    def row(self, vertex):
        return self._rindex.get(vertex, -1)
    def rows(self):
        return self._rindex
    def set_row(self, vertex, r):
        if r > self._maxr: self._maxr = r
        self._rindex[vertex] = r

    def is_ground(self, vertex):
        return vertex in self._grounds
    def grounds(self):
        return self._grounds
    def set_ground(self, vertex, flag=True):
        if flag:
            self._grounds.add(vertex)
        else:
            self._grounds.discard(vertex)

    def clear_vdata(self, vertex):
        if vertex in self._vdata:
            del self._vdata[vertex]
    def vdata_keys(self, vertex):
        return self._vdata.get(vertex, {}).keys()
    def vdata(self, vertex, key, default=None):
        if vertex in self._vdata:
            return self._vdata[vertex].get(key,default)
        else:
            return default
    def set_vdata(self, vertex, key, val):
        if vertex in self._vdata:
            self._vdata[vertex][key] = val
        else:
            self._vdata[vertex] = {key:val}

    def clear_edata(self, edge):
        self._edata.pop(edge, None)

    def edata_keys(self, edge):
        return self._edata.get(edge, {}).keys()

    def edata(self, edge, key, default=None):
        return self._edata.get(edge, {}).get(key, default)

    def set_edata(self, edge, key, val):
        if edge in self._edata:
            self._edata[edge][key] = val
        else:
            self._edata[edge] = {key: val}

    def edata_dict(self, edge):
        return dict(self._edata.get(edge, {}))

    def set_edata_dict(self, edge, d):
        self.clear_edata(edge)
        if d:
            self._edata[edge] = dict(d)

    @classmethod
    def from_json(cls, js:Union[str,Dict[str,Any]]) -> 'Multigraph':
        """Converts the given .qgraph json string into a Multigraph.
        Works with the output of :meth:`to_json`."""
        from .jsonparser import json_to_graph
        return json_to_graph(js,backend='multigraph') # type:ignore


# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import copy
import unittest
import sys
from typing import List, Optional

if __name__ == "__main__":
    sys.path.append("..")
    sys.path.append(".")
import numpy as np

from pyzx_param.linalg import Mat2, MatLike
from pyzx_param.routing.cnot_mapper import (
    CostMetric,
    gauss,
    ElimMode,
    FitnessFunction,
    sequential_gauss,
)
from pyzx_param.routing.architecture import (
    Architecture,
    architectures,
    dynamic_size_architectures,
    hamiltonian_path_architectures,
    create_architecture,
    SQUARE,
    FULLY_CONNECTED,
    IBMQ_SINGAPORE,
)
from pyzx_param.routing.parity_maps import CNOT_tracker
from pyzx_param.routing.machine_learning import GeneticAlgorithm
from pyzx_param.circuit import CNOT
from pyzx_param.extract import permutation_as_swaps
from pyzx_param.generate import build_random_parity_map
from pyzx_param.routing.steiner import steiner_gauss

SEED = 42


class TestSteiner(unittest.TestCase):
    def setUp(self) -> None:
        np.random.seed(SEED)
        self.n_tests = 1
        self.arch = create_architecture(
            SQUARE, n_qubits=9
        )  # Needs to have a square number of qubits to test the square architecture.
        self.n_qubits = self.arch.n_qubits
        depth = 16
        self.circuit = [CNOT_tracker(self.arch.n_qubits) for _ in range(self.n_tests)]
        self.matrix: List[Mat2] = [
            Mat2(build_random_parity_map(self.n_qubits, depth, self.circuit[i]))
            for i in range(self.n_tests)
        ]
        self.aggr_circ = CNOT_tracker(self.arch.n_qubits)
        for c in self.circuit:
            for g in c.gates:
                self.aggr_circ.add_gate(g)

    def assertGates(self, circuit, architecture=None):
        if architecture is None:
            architecture = self.arch
        for gate in circuit.gates:
            if hasattr(gate, "name") and gate.name == "CNOT":
                qubits = (gate.control, gate.target)
                edge = (
                    architecture.qubit2vertex(qubits[0]),
                    architecture.qubit2vertex(qubits[1]),
                )
                edges = list(architecture.graph.edges())
                edges += [tuple(reversed(edge)) for edge in edges]
                self.assertIn(edge, edges)

    def assertMat2Equal(self, m1: Mat2, m2: Mat2, triangle: bool = False):
        if triangle:
            self.assertListEqual(
                *[
                    [
                        m.data[i][j]
                        for i in range(self.n_qubits)
                        for j in range(self.n_qubits)
                        if i >= j
                    ]
                    for m in [m1, m2]
                ]
            )
        else:
            self.assertListEqual(m1.data, m2.data)

    def assertCircuitEquivalentMat2(self, circuit: CNOT_tracker, matrix: Mat2):
        self.assertMat2Equal(circuit.matrix, matrix)

    def assertCircuitEquivalent(self, c1: CNOT_tracker, c2: CNOT_tracker):
        self.assertMat2Equal(c1.matrix, c2.matrix)

    def assertIsPerm(self, l: List[int]):
        self.assertTrue(all([i in l for i in range(len(l))]))

    def test_all_cnots_valid(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                circuit = CNOT_tracker(self.arch.n_qubits)
                matrix = self.matrix[i].copy()
                _ = gauss(
                    ElimMode.STEINER_MODE,
                    matrix,
                    architecture=self.arch,
                    full_reduce=True,
                    y=circuit,
                )
                self.assertGates(circuit)

    def test_architectures(self):
        for i in range(self.n_tests):
            for name in architectures:
                # print("Testing", name)
                if name in dynamic_size_architectures:
                    if name not in hamiltonian_path_architectures:
                        # continue
                        # This is DENSITY and it gets in an infinite loop when building steiner trees!
                        architecture = create_architecture(name, n_qubits=self.n_qubits)
                    else:
                        try:
                            architecture = create_architecture(
                                name, n_qubits=self.n_qubits
                            )
                        except KeyError as e:
                            if name == SQUARE:
                                print(
                                    "WARNING! skipping SQUARE architecture due to non-square number of qubits:",
                                    self.n_qubits,
                                )
                                continue
                            else:
                                raise e
                elif name in hamiltonian_path_architectures:
                    architecture = create_architecture(name)
                else:
                    architecture = create_architecture(name)
                if self.n_qubits == architecture.n_qubits:
                    with self.subTest(i=i, arch=architecture.name):
                        self.do_gauss(
                            ElimMode.STEINER_MODE,
                            self.matrix[i],
                            architecture=architecture,
                        )

    def test_gauss_small(self):
        n_qubits = 2
        architecture = create_architecture(FULLY_CONNECTED, n_qubits=n_qubits)
        matrix = Mat2([[1, 1], [1, 0]])
        for mode in [ElimMode.GAUSS_MODE, ElimMode.STEINER_MODE]:
            with self.subTest(i=mode):
                mtx = matrix.copy()
                circuit = CNOT_tracker(n_qubits)
                _ = gauss(
                    mode,
                    mtx,
                    architecture=architecture,
                    full_reduce=False,
                    y=circuit,
                )
                self.assertGates(circuit)
                self.assertMat2Equal(circuit.matrix * matrix, mtx)

    def do_gauss(
        self,
        mode: ElimMode,
        mat: Mat2,
        full_reduce: bool = True,
        with_assert: bool = True,
        architecture: Optional[Architecture] = None,
    ):
        if architecture is None:
            architecture = self.arch
        circuit = CNOT_tracker(self.arch.n_qubits)
        matrix = mat.copy()
        rank = gauss(
            mode, matrix, architecture=architecture, full_reduce=full_reduce, y=circuit
        )
        if with_assert and mode == ElimMode.STEINER_MODE:
            self.assertGates(circuit, architecture)
        if with_assert and full_reduce:
            self.assertMat2Equal(circuit.matrix, mat)
        return circuit, matrix, rank

    def test_gauss(self) -> None:
        for i in range(self.n_tests):
            with self.subTest(i=i):
                for full_reduce in [False, True]:
                    matrices: List[Mat2] = []
                    ranks = []
                    circuits = []
                    for mode in [ElimMode.GAUSS_MODE, ElimMode.STEINER_MODE]:
                        circuit, matrix, rank = self.do_gauss(
                            mode,
                            self.matrix[i],
                            full_reduce=full_reduce,
                            with_assert=True,
                        )
                        circuits.append(circuit)
                        ranks.append(rank)
                        matrices.append(matrix)
                        if full_reduce:
                            self.assertCircuitEquivalent(circuit, self.circuit[i])
                    # self.assertEqual(*ranks)
                    if full_reduce:
                        self.assertMat2Equal(matrices[0], matrices[1])
                        self.assertCircuitEquivalent(circuits[0], circuits[1])
                    else:
                        self.assertMat2Equal(matrices[0], matrices[1], triangle=True)

    def test_initial_circuit(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                self.assertCircuitEquivalentMat2(self.circuit[i], self.matrix[i])

    def reverse_permutation(self, perm):
        return [perm.tolist().index(i) for i in range(len(perm))]

    def do_permuted_gauss(
        self,
        matrix: Mat2,
        perm_rows: List[int],
        perm_columns: List[int],
        mode=ElimMode.STEINER_MODE,
        full_reduce: bool = True,
        with_assert: bool = True,
    ):
        if with_assert:
            self.assertIsPerm(perm_rows)
            self.assertIsPerm(perm_columns)
        map = matrix.copy()
        map.permute_rows(perm_rows)
        map.permute_cols(perm_columns)
        circuit, map, rank = self.do_gauss(
            mode, map, full_reduce=full_reduce, with_assert=with_assert
        )
        return circuit, map, rank

    def test_permuted_gauss(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                perm_rows = list(np.random.permutation(self.n_qubits))
                perm_columns = list(np.random.permutation(self.n_qubits))
                self.do_permuted_gauss(self.matrix[i], perm_rows, perm_columns)

    def test_genetic_optimization(self):
        for i in range(1):  # Takes too long otherwise
            with self.subTest(i=i):
                population = 4
                crossover_prob = 0.8
                mutate_prob = 0.2
                n_iter = 4
                optimizer = GeneticAlgorithm(
                    population,
                    crossover_prob,
                    mutate_prob,
                    FitnessFunction(CostMetric.COUNT, self.matrix[i], ElimMode.STEINER_MODE, self.arch),
                )
                best_permutation = optimizer.find_optimum(self.n_qubits, n_iter)
                self.do_permuted_gauss(
                    self.matrix[i], best_permutation, best_permutation
                )

    def test_pso_optimization(self):
        modes = [
            ElimMode.STEINER_MODE,
            ElimMode.GENETIC_STEINER_MODE,
            ElimMode.PSO_STEINER_MODE,
        ]
        for p in range(1):  # Takes too long otherwise
            with self.subTest(i=p):
                order = np.random.permutation(self.n_tests)[: p + 1]
                matrices = [self.matrix[c] for c in order]
                for mode in modes:
                    for permute_input, permute_output in [
                        (i, j) for i in [True, False] for j in [True, False]
                    ]:
                        with self.subTest(i=(p, mode, permute_input, permute_output)):
                            circuits, perms, score = sequential_gauss(
                                [m.copy() for m in matrices],  # type: ignore # MatLike should be a valid ArrayLike
                                mode=mode,
                                architecture=self.arch,
                                full_reduce=True,
                                # Small values to keep the test fast
                                # It doesn't need to find an optimized solution, it only needs to do a non-trivial run
                                n_steps=4,
                                swarm_size=4,
                                population_size=4,
                                n_iterations=1,
                                input_perm=permute_input,
                                output_perm=permute_output,
                            )
                            if not permute_input:
                                self.assertListEqual(
                                    perms[0], [i for i in range(self.n_qubits)]
                                )
                            if not permute_output:
                                self.assertListEqual(
                                    perms[-1],
                                    [i for i in range(self.n_qubits)],
                                )
                            aggr_c = CNOT_tracker(self.n_qubits)
                            for circ in circuits:
                                for gate in circ.gates:
                                    aggr_c.add_gate(gate)

                            aggr_c2 = CNOT_tracker(self.n_qubits)
                            # Undo the initial permutation
                            for q1, q2 in permutation_as_swaps(
                                {k: v for k, v in enumerate(perms[0])}
                            ):
                                aggr_c2.add_gate(CNOT(q1, q2))
                                aggr_c2.add_gate(CNOT(q2, q1))
                                aggr_c2.add_gate(CNOT(q1, q2))
                            for circ in [self.circuit[i] for i in order]:
                                for gate in circ.gates:
                                    aggr_c2.add_gate(gate)
                            # Undo the initial permutation
                            for q1, q2 in permutation_as_swaps(
                                {v: k for k, v in enumerate(perms[-1])}
                            ):
                                aggr_c2.add_gate(CNOT(q1, q2))
                                aggr_c2.add_gate(CNOT(q2, q1))
                                aggr_c2.add_gate(CNOT(q1, q2))
                            self.assertCircuitEquivalent(aggr_c2, aggr_c)
                            for i in range(p):
                                with self.subTest(i=(i, mode)):
                                    # Check if all gates are allowed
                                    self.assertGates(circuits[i])
                                    # Check if the circuit is equivalent to the original matrix
                                    self.assertCircuitEquivalentMat2(
                                        circuits[i],
                                        matrices[i],
                                    )
                                    # Check if the circuit is equivalent to the extracted circuit given the optimized permutations
                                    c, _, _ = self.do_permuted_gauss(
                                        matrices[i].copy(), perms[i + 1], perms[i]  # type: ignore # MatLike should be a valid ArrayLike
                                    )
                                    self.assertCircuitEquivalent(circuits[i], c)
                                    # Check if their metrics are the same.
                                    self.assertEqual(
                                        circuits[i].count_cnots(), c.count_cnots()
                                    )
                                    self.assertEqual(
                                        circuits[i].cnot_depth(), c.cnot_depth()
                                    )

    @unittest.skip("This test fails because the steiner_gauss tries to find a steiner tree in a disconnected subgraph of the architecture")
    def test_small_steiner_gauss(self):
        """
        Small regression test for steiner_gauss
        """
        arch = create_architecture(IBMQ_SINGAPORE)

        print("Architecture qubit mapping")
        print(arch.qubit_map)

        n_qubits = arch.n_qubits
        c = CNOT_tracker(n_qubits)
        cnots = [
            (4, 17),
            (7, 10),
            (11, 9),
            (2, 14),
            (6, 19),
            (6, 11),
            (2, 11),
            (12, 19),
            (6, 4),
            (2, 0),
            (5, 19),
            (5, 15),
            (16, 1),
            (5, 11),
            (3, 18),
            (12, 11),
        ]
        for u, v in cnots:
            c.add_gate(CNOT(u, v))

        tracker = CNOT_tracker(n_qubits)
        steiner_gauss(c.matrix, arch, full_reduce=True, y=tracker)
        self.assertGates(tracker, architecture=arch)


if __name__ == "__main__":
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains the definition of commonly used
quantum gates for use in the Circuit class.
"""

import copy
import math
from fractions import Fraction
from typing import Dict, List, Optional, Type, ClassVar, TypeVar, Generic, Set

from ..utils import EdgeType, VertexType, FractionLike
from ..graph.base import BaseGraph, VT, ET

# We need this type variable so that the subclasses of Gate return the correct type for functions like copy()
Tvar = TypeVar('Tvar', bound='Gate')

class TargetMapper(Generic[VT]):
    """
    This class is used to map the target parameters of a gate to rows, qubits, and vertices
    when converting them into a graph. Used by :func:`~pyzx.circuit.gates.Gate.to_graph`.
    """
    _qubits: Dict[int, int]
    _rows: Dict[int, int]
    _prev_vs: Dict[int, VT]
    _labels: Set[int]
    _max_row: int

    def __init__(self):
        self._qubits = {}
        self._rows = {}
        self._prev_vs = {}
        self._labels = set()
        self._max_row = 0

    def labels(self) -> Set[int]:
        """
        Returns the mapped labels.
        """
        return self._labels

    def to_qubit(self, l: int) -> int:
        """
        Maps a label to the qubit id in the graph.
        """
        return self._qubits[l]

    def set_qubit(self, l: int, q: int) -> None:
        """
        Sets the qubit id for a label.
        """
        self._qubits[l] = q

    def next_row(self, l: int) -> int:
        """
        Returns the next free row in the label's qubit.
        """
        return self._rows[l]

    def next_row_or_default(self, l: int, default: int) -> int:
        """
        Returns the next free row in the label's qubit if it is tracked.
        Otherwise, the default value is returned.
        """
        return self._rows.get(l, default)

    def set_next_row(self, l: int, row: int) -> None:
        """
        Sets the next free row in the label's qubit.
        """
        self._rows[l] = row
        self._max_row = max(self._max_row, row)

    def advance_next_row(self, l: int) -> None:
        """
        Advances the next free row in the label's qubit by one.
        """
        self.set_next_row(l, self.next_row(l)+1)

    def shift_all_rows(self, n: int) -> None:
        """
        Shifts all 'next rows' by n.
        """
        for l in self._rows.keys():
            self._rows[l] += n
        self._max_row += n

    # def set_all_rows(self, n: int) -> None:
    #     """
    #     Set the value of all 'next rows'.
    #     """
    #     for l in self._rows.keys():
    #         self._rows[l] = n
    #     self._max_row = max(self._max_row, n)

    def set_all_rows_to_max(self) -> None:
        for l in self._rows.keys():
            self._rows[l] = self._max_row

    def set_max_row(self, max_row: int) -> None:
        self._max_row = max_row

    def max_row(self) -> int:
        """
        Returns the highest 'next row' number.
        """
        return self._max_row

    def prev_vertex(self, l: int) -> VT:
        """
        Returns the previous vertex in the label's qubit.
        """
        return self._prev_vs[l]

    def set_prev_vertex(self, l: int, v: VT) -> None:
        """
        Sets the previous vertex in the label's qubit.
        """
        self._prev_vs[l] = v

    def add_label(self, l: int, row: int) -> None:
        """
        Adds a tracked label.

        :raises: ValueError if the label is already tracked.
        """
        if l in self._labels:
            raise ValueError("Label {} already in use".format(str(l)))
        q = self._qubits.get(l, len(self._qubits))
        self.set_qubit(l, q)
        self.set_next_row(l, row)
        self._labels.add(l)
        # r = self.max_row()
        # self.set_all_rows_to_max()

        # if compress_rows:
        #     self.set_next_row(l, r)
        # else:
        #     self.set_next_row(l, r+1)

    def remove_label(self, l: int) -> None:
        """
        Removes a tracked label.

        :raises: ValueError if the label is not tracked.
        """
        if l not in self._labels:
            raise ValueError("Label {} not in use".format(str(l)))
        
        # self.set_all_rows_to_max()
        # if not compress_rows:
        #     self.shift_all_rows(1)

        self._labels.remove(l)

class Gate(object):
    """Base class for representing quantum gates."""
    name:               ClassVar[str] = "BaseGate"
    qasm_name:          ClassVar[str] = 'undefined'
    qasm_name_adjoint:  ClassVar[str] = 'undefined'
    qc_name:            ClassVar[str] = 'undefined'
    quipper_name:       ClassVar[str] = 'undefined'
    print_phase:        ClassVar[bool] = False
    index = 0
    def __str__(self) -> str:
        attribs = []
        if hasattr(self, "control"): attribs.append(str(self.control))
        if hasattr(self, "target"): attribs.append(str(self.target))
        if self.print_phase:
            if hasattr(self, "phase"):
                attribs.append("phase={!s}".format(self.phase))
            elif hasattr(self, "phases"):
                attribs.append("phases={}".format(",".join(str(p) for p in self.phases)))
        return "{}{}({})".format(
                        self.name,
                        ("*" if (hasattr(self,"adjoint") and self.adjoint) else ""),
                        ",".join(attribs))

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other: object) -> bool:
        if type(self) != type(other): return False
        for a in ["target","control","phase","adjoint"]:
            if hasattr(self,a):
                if not hasattr(other,a): return False
                if getattr(self,a) != getattr(other,a): return False
            elif hasattr(other,a): return False
        assert isinstance(other, Gate)
        if self.index != other.index: return False
        return True

    def _max_target(self) -> int:
        qubits = self.target        # type: ignore # due to ParityPhase
        if hasattr(self, "control"):
            qubits = max([qubits, self.control])
        return qubits

    def __add__(self, other):
        from . import Circuit
        c = Circuit(self._max_target()+1)
        c.add_gate(self)
        c += other
        return c

    def __matmul__(self,other):
        from . import Circuit
        c = Circuit(self._max_target()+1)
        c.add_gate(self)
        c2 = Circuit(other._max_target()+1)
        c2.add_gate(other)
        return c@c2

    def copy(self: Tvar) -> Tvar:
        return copy.copy(self)

    def to_adjoint(self: Tvar) -> Tvar:
        g = self.copy()
        if hasattr(g, "phase"):
            g.phase = -g.phase
        if hasattr(g, "adjoint"):
            g.adjoint = not g.adjoint
        return g

    def tcount(self) -> int:
        return 0

    def reposition(self: Tvar, mask: List[int], bit_mask: Optional[List[int]] = None) -> Tvar:
        g = self.copy()
        if hasattr(g, "target"):
            g.target = mask[g.target]
        if hasattr(g, "control"):
            g.control = mask[g.control]
        return g

    def to_basic_gates(self) -> List['Gate']:
        return [self]

    def to_quipper(self) -> str:
        n = self.quipper_name
        if n == 'undefined':
            bg = self.to_basic_gates()
            if len(bg) == 1:
                raise TypeError("Gate {} doesn't have a Quipper description".format(str(self)))
            return "\n".join(g.to_quipper() for g in bg)
        s = 'QGate["{}"]{}({!s})'.format(n,
                                         ("*" if (hasattr(self,"adjoint") and self.adjoint) else ""),
                                         self.target) # type: ignore # due to ParityPhase
        if hasattr(self, "control"):
            s += ' with controls=[+{!s}]'.format(self.control)
        s += ' with nocontrol'
        return s

    def to_qasm(self) -> str:
        n = self.qasm_name
        if n == 'undefined':
            bg = self.to_basic_gates()
            if len(bg) == 1:
                raise TypeError("Gate {} doesn't have a QASM description".format(str(self)))
            return "\n".join(g.to_qasm() for g in bg)
        if hasattr(self, "adjoint") and self.adjoint:
            n = self.qasm_name_adjoint

        args = []
        for a in ["ctrl1","ctrl2", "control", "target"]:
            if hasattr(self, a): args.append("q[{:d}]".format(getattr(self,a)))
        param = ""
        if self.print_phase:
            if hasattr(self, "phase"):
                param = "({}*pi)".format(float(self.phase))
            elif hasattr(self, "phases"):
                param = "({})".format(",".join("{}*pi".format(float(p)) for p in self.phases))
        return "{}{} {};".format(n, param, ", ".join(args))

    def to_qc(self) -> str:
        n = self.qc_name
        if hasattr(self, "adjoint") and self.adjoint:
            n += "*"
        if n == 'undefined':
            if isinstance(self, (ZPhase, XPhase)):
                bg = self.split_phases()
                if any(g.qc_name == 'undefined' for g in bg):
                    raise TypeError("Gate {} doesn't have a .qc description".format(str(self)))
            else:
                bg = self.to_basic_gates()
                if len(bg) == 1:
                    raise TypeError("Gate {} doesn't have a .qc description".format(str(self)))
            return "\n".join(g.to_qc() for g in bg)
        args = []
        for a in ["ctrl1","ctrl2", "control", "target"]:
            if hasattr(self, a): args.append("q{:d}".format(getattr(self,a)))

        # if hasattr(self, "phase") and self.print_phase:
        #     args.insert(0, phase_to_s(self.phase))
        return "{} {}".format(n, " ".join(args))

    def to_graph(self, g: BaseGraph[VT,ET], q_mapper: TargetMapper[VT], c_mapper: TargetMapper[VT]) -> None:
        """
        Add the converted gate to the graph.

        :param g: The graph to add the gate to.
        :param q_mapper: A mapper for qubit labels.
        :param c_mapper: A mapper for bit labels.
        """
        raise NotImplementedError("to_graph() must be implemented by each Gate subclass.")

    def graph_add_node(self,
                g: BaseGraph[VT,ET],
                mapper: TargetMapper[VT],
                t: VertexType,
                l: int, r: int,
                phase: FractionLike=0,
                etype: EdgeType=EdgeType.SIMPLE,
                ground: bool = False) -> VT:
        v = g.add_vertex(t, mapper.to_qubit(l), r, phase, ground)
        g.add_edge((mapper.prev_vertex(l), v), etype)
        mapper.set_prev_vertex(l, v)
        return v

class ZPhase(Gate):
    name = 'ZPhase'
    qasm_name = 'rz'
    quipper_name = 'ZPhase'
    print_phase = True
    def __init__(self, target: int, phase: FractionLike) -> None:
        self.target = target
        self.phase = phase

    def to_graph(self, g, q_mapper, _c_mapper):
        self.graph_add_node(g,q_mapper, VertexType.Z, self.target, q_mapper.next_row(self.target), self.phase)
        q_mapper.advance_next_row(self.target)

    def to_quipper(self):
        if not self.print_phase:
            return super().to_quipper()
        return 'QRot["exp(-i%Z)",{!s}]({!s})'.format(math.pi*self.phase/2,self.target)

    def to_emoji(self,strings: List[List[str]]) -> None:
        s = ''
        phase = self.phase % 2
        if phase == Fraction(1,2): s = ':Zp4:'
        elif phase == Fraction(1,4): s = ':Zp4:'
        elif phase == Fraction(1,1): s = ':Zp:'
        elif phase == Fraction(3,4): s = ':Z3p4:'
        elif phase == Fraction(5,4): s = ':Z5p4:'
        elif phase == Fraction(3,2): s = ':Z3p2:'
        elif phase == Fraction(7,4): s = ':Z7p4:'
        else: raise Exception("Unsupported phase " + str(phase))
        strings[self.target].append(s)

    def tcount(self):
        return 1 if self.phase.denominator > 2 else 0

    def split_phases(self) -> List['ZPhase']:
        if not self.phase: return []
        if self.phase == 1: return [Z(self.target)]
        if self.phase.denominator == 2:
            if self.phase.numerator % 4 == 1:
                return [S(self.target)]
            else: return [S(self.target, adjoint=True)]
        elif self.phase.denominator == 4:
            gates: List['ZPhase'] = []
            n = self.phase.numerator % 8
            if n == 3 or n == 5:
                gates.append(Z(self.target))
                n = (n-4)%8
            if n == 1: gates.append(T(self.target))
            if n == 7: gates.append(T(self.target, adjoint=True))
            return gates
        else:
            return [self]


class Z(ZPhase):
    name = 'Z'
    qasm_name = 'z'
    qc_name = 'Z'
    quipper_name = 'Z'
    print_phase = False
    def __init__(self, target: int) -> None:
        super().__init__(target, Fraction(1,1))

class S(ZPhase):
    name = 'S'
    qasm_name = 's'
    qasm_name_adjoint = 'sdg'
    qc_name = 'S'
    quipper_name = 'S'
    print_phase = False
    def __init__(self, target: int, adjoint:bool=False) -> None:
        super().__init__(target, Fraction(1,2)*(-1 if adjoint else 1))
        self.adjoint = adjoint

class T(ZPhase):
    name = 'T'
    qasm_name = 't'
    qasm_name_adjoint = 'tdg'
    qc_name = 'T'
    quipper_name = 'T'
    print_phase = False
    def __init__(self, target: int, adjoint:bool=False) -> None:
        super().__init__(target, Fraction(1,4)*(-1 if adjoint else 1))
        self.adjoint = adjoint

class XPhase(Gate):
    name = 'XPhase'
    qasm_name = 'rx'
    quipper_name = 'XPhase'
    print_phase = True
    def __init__(self, target: int, phase: FractionLike=0) -> None:
        self.target = target
        self.phase = phase

    def to_graph(self, g, q_mapper, _c_mapper):
        self.graph_add_node(g, q_mapper, VertexType.X, self.target, q_mapper.next_row(self.target), self.phase)
        q_mapper.advance_next_row(self.target)

    def to_emoji(self,strings: List[List[str]]) -> None:
        s = ''
        phase = self.phase % 2
        if phase == Fraction(1,2): s = ':Xp4:'
        elif phase == Fraction(1,4): s = ':Xp4:'
        elif phase == Fraction(1,1): s = ':Xp:'
        elif phase == Fraction(3,4): s = ':X3p4:'
        elif phase == Fraction(5,4): s = ':X5p4:'
        elif phase == Fraction(3,2): s = ':X3p2:'
        elif phase == Fraction(7,4): s = ':X7p4:'
        else: raise Exception("Unsupported phase " + str(phase))
        strings[self.target].append(s)

    def to_quipper(self):
        if not self.print_phase:
            return super().to_quipper()
        return 'QRot["exp(-i%X)",{!s}]({!s})'.format(math.pi*self.phase/2,self.target)

    def tcount(self):
        return 1 if self.phase.denominator > 2 else 0

    def split_phases(self) -> List[Gate]:
        if not self.phase: return []
        if self.phase == 1: return [NOT(self.target)]
        gates: List[Gate] = [HAD(self.target)]
        if self.phase.denominator == 2:
            if self.phase.numerator % 4 == 1:
                gates.append(S(self.target))
            else: gates.append(S(self.target, adjoint=True))
        elif self.phase.denominator == 4:
            n = self.phase.numerator % 8
            if n == 3 or n == 5:
                gates.append(Z(self.target))
                n = (n-4)%8
            if n == 1: gates.append(T(self.target))
            if n == 7: gates.append(T(self.target, adjoint=True))
        else:
            gates.append(ZPhase(self.target, self.phase))
        gates.append(HAD(self.target))
        return gates

class SX(XPhase):
    name = 'SX'
    qasm_name = 'sx'
    qasm_name_adjoint = 'sxdg'
    print_phase = False
    def __init__(self, target: int, adjoint:bool=False) -> None:
        super().__init__(target, Fraction(1,2)*(-1 if adjoint else 1))
        self.adjoint = adjoint

class CSX(Gate):
    name = 'CSX'
    qasm_name = 'csx'
    def __init__(self, control: int, target: int) -> None:
        self.target = target
        self.control = control

    def to_basic_gates(self):
        return [HAD(self.target)] + \
               CPhase(self.control,self.target,Fraction(1,2)).to_basic_gates() + \
               [HAD(self.target)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class YPhase(Gate):
    name = 'YPhase'
    qasm_name = 'ry'
    quipper_name = 'YPhase'
    print_phase = True
    def __init__(self, target: int, phase: FractionLike=0) -> None:
        self.target = target
        self.phase = phase

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, YPhase): return False
        if self.index != other.index: return False
        if self.target == other.target and self.phase == other.phase:
            return True
        return False

    def __str__(self) -> str:
        return 'QRot["exp(-i%Y)",{!s}]({!s})'.format(math.pi*self.phase/2,self.target)

    def to_basic_gates(self):
        return [ZPhase(self.target, Fraction(1,2)), XPhase(self.target, -self.phase), ZPhase(self.target, -Fraction(1,2))]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

    def tcount(self):
        return 1 if self.phase.denominator > 2 else 0

class Y(YPhase):
    name = 'Y'
    qasm_name = 'y'
    qc_name = 'Y'
    quipper_name = 'Y'
    print_phase = False
    def __init__(self, target: int) -> None:
        super().__init__(target, phase = Fraction(1,1))

class CY(Gate):
    name = 'CY'
    qasm_name = 'cy'
    def __init__(self, control: int, target: int) -> None:
        self.target = target
        self.control = control

    def to_basic_gates(self):
        return [S(self.target,adjoint=True),
                CNOT(self.control,self.target),
                S(self.target)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class NOT(XPhase):
    name = 'NOT'
    qasm_name = 'x'
    qc_name = 'X'
    quipper_name = 'not'
    print_phase = False
    def __init__(self, target: int) -> None:
        super().__init__(target, phase = Fraction(1,1))

class HAD(Gate):
    name = 'HAD'
    qasm_name = 'h'
    qc_name = 'H'
    quipper_name = 'H'
    def __init__(self, target: int) -> None:
        self.target = target

    def to_graph(self, g, q_mapper, _c_mapper):
        v = g.add_vertex(VertexType.Z, q_mapper.to_qubit(self.target), q_mapper.next_row(self.target))
        g.add_edge((q_mapper.prev_vertex(self.target),v), EdgeType.HADAMARD)
        q_mapper.set_prev_vertex(self.target, v)
        q_mapper.advance_next_row(self.target)

    def to_emoji(self,strings: List[List[str]]) -> None:
        strings[self.target].append(':H_:')

class CNOT(Gate):
    name = 'CNOT'
    qasm_name = 'cx'
    qc_name = 'Tof'
    quipper_name = 'not'
    def __init__(self, control: int, target: int) -> None:
        self.target = target
        self.control = control
    def to_graph(self, g, q_mapper, _c_mapper):
        [top, bot] = sorted([self.target, self.control])
        r = max(q_mapper.next_row(r_) for r_ in range(top, bot + 1))
        t = self.graph_add_node(g, q_mapper, VertexType.X, self.target, r)
        c = self.graph_add_node(g, q_mapper, VertexType.Z, self.control, r)
        g.add_edge((t,c))
        for r_ in range(top, bot + 1):
          q_mapper.set_next_row(r_, r+1)
        g.scalar.add_power(1)

    def to_emoji(self,strings: List[List[str]]) -> None:
        c,t = self.control, self.target
        mi = min([c,t])
        ma = max([c,t])
        r = max([len(s) for s in strings[mi:ma+1]])
        for s in (strings[mi:ma+1]):
                s.extend([':W_:']*(r-len(s)))
        for i in range(mi+1,ma): strings[i].append(':Wud:')
        if c<t:
            strings[self.control].append(':Zd:')
            strings[self.target].append(':Xu:')
        else:
            strings[self.control].append(':Zu:')
            strings[self.target].append(':Xd:')

class CZ(Gate):
    name = 'CZ'
    qasm_name = 'cz'
    qc_name = 'Z'
    quipper_name = 'Z'
    def __init__(self, control: int, target: int) -> None:
        self.target = target
        self.control = control
    def __eq__(self,other: object) -> bool:
        if not isinstance(other, CZ): return False
        if not type(self) == type(other): return False
        if self.index != other.index: return False
        if (
            (self.target == other.target and self.control == other.control) or
            (self.target == other.control and self.control == other.target)):
            return True
        return False

    def to_graph(self, g, q_mapper, _c_mapper):
        r = max(q_mapper.next_row(self.target), q_mapper.next_row(self.control))
        t = self.graph_add_node(g, q_mapper, VertexType.Z, self.target, r)
        c = self.graph_add_node(g, q_mapper, VertexType.Z, self.control, r)
        g.add_edge((t,c), EdgeType.HADAMARD)
        q_mapper.set_next_row(self.target, r+1)
        q_mapper.set_next_row(self.control, r+1)
        g.scalar.add_power(1)

    def to_emoji(self,strings: List[List[str]]) -> None:
        c,t = self.control, self.target
        strings[t].append(':H_:')
        CNOT(c,t).to_emoji(strings)
        strings[t].append(':H_:')


class XCX(CZ):
    '''This class represents the X-controlled-X gate.'''
    name = 'XCX'
    qasm_name = 'undefined'
    qc_name = 'undefined'
    quipper_name = 'X'
    def to_graph(self, g, q_mapper, _c_mapper):
        r = max(q_mapper.next_row(self.target), q_mapper.next_row(self.control))
        t = self.graph_add_node(g, q_mapper, VertexType.X, self.target, r)
        c = self.graph_add_node(g, q_mapper, VertexType.X, self.control, r)
        g.add_edge((t,c), EdgeType.HADAMARD)
        q_mapper.set_next_row(self.target, r+1)
        q_mapper.set_next_row(self.control, r+1)
        g.scalar.add_power(1)

    def to_basic_gates(self):
        return [HAD(self.control), CNOT(self.control,self.target), HAD(self.control)]

class SWAP(CZ):
    name = 'SWAP'
    qasm_name = 'swap'
    qc_name = 'undefined'
    quipper_name = 'undefined'
    def to_basic_gates(self):
        c1 = CNOT(self.control, self.target)
        c2 = CNOT(self.target, self.control)
        return [c1,c2,c1]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class CRX(Gate):
    name = 'CRX'
    qasm_name = 'crx'
    print_phase = True
    def __init__(self, control: int, target: int, phase: FractionLike) -> None:
        self.target = target
        self.control = control
        self.phase = phase

    def to_basic_gates(self):
        return [ZPhase(self.target,Fraction(1,2)),
                CNOT(self.control,self.target)] + \
               U3(self.target,-self.phase/2,0,0).to_basic_gates() + \
               [CNOT(self.control,self.target)] + \
               U3(self.target,self.phase/2,Fraction(-1,2),0).to_basic_gates()

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class CRY(Gate):
    name = 'CRY'
    qasm_name = 'cry'
    print_phase = True
    def __init__(self, control: int, target: int, phase: FractionLike) -> None:
        self.target = target
        self.control = control
        self.phase = phase

    def to_basic_gates(self):
        return YPhase(self.target,self.phase/2).to_basic_gates() + \
               [CNOT(self.control,self.target)] + \
               YPhase(self.target,-self.phase/2).to_basic_gates() + \
               [CNOT(self.control,self.target)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class CRZ(Gate):
    name = 'CRZ'
    qasm_name = 'crz'
    print_phase = True
    def __init__(self, control: int, target: int, phase: FractionLike) -> None:
        self.target = target
        self.control = control
        self.phase = phase

    def to_basic_gates(self):
        phase1 = self.phase / 2
        phase2 = -self.phase / 2
        try:
            phase1 = Fraction(phase1) % 2
            phase2 = Fraction(phase2) % 2
        except Exception:
            pass
        return [ZPhase(self.target, phase1),
                CNOT(self.control, self.target),
                ZPhase(self.target, phase2),
                CNOT(self.control, self.target)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class RXX(Gate):
    name = 'RXX'
    qasm_name = 'rxx'
    print_phase = True
    def __init__(self, control: int, target: int, phase: FractionLike) -> None:
        self.target = target
        self.control = control
        self.phase = phase

    def to_basic_gates(self):
        return U3(self.control,Fraction(1,2),self.phase,0).to_basic_gates() + \
               [HAD(self.target),
                CNOT(self.control,self.target),
                ZPhase(self.target,-self.phase),
                CNOT(self.control,self.target),
                HAD(self.target)] + \
               U2(self.control,Fraction(-1,1),(Fraction(1,1)-self.phase)%2).to_basic_gates()

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class CPhase(CRZ):
    name = 'CPhase'
    qasm_name = 'cp'

    def to_basic_gates(self):
        phase1 = self.phase / 2
        phase2 = -self.phase / 2
        try:
            phase1 = Fraction(phase1) % 2
            phase2 = Fraction(phase2) % 2
        except Exception:
            pass
        return [ZPhase(self.control, phase1),
                CNOT(self.control, self.target),
                ZPhase(self.target, phase2),
                CNOT(self.control, self.target),
                ZPhase(self.target, phase1)]

class CHAD(Gate):
    name = 'CHAD'
    qasm_name = 'ch'
    quipper_name = 'undefined'

    def __init__(self, control: int, target: int) -> None:
        self.target = target
        self.control = control

    def to_basic_gates(self):
        return [HAD(self.target),S(self.target,adjoint=True),
                CNOT(self.control,self.target),
                HAD(self.target),T(self.target),
                CNOT(self.control,self.target),
                T(self.target),HAD(self.target),S(self.target),NOT(self.target),S(self.control)]


    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class ParityPhase(Gate):
    name = 'ParityPhase'
    print_phase = True
    def __init__(self, phase: FractionLike, *targets: int):
        self.targets = targets
        self.phase = phase

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ParityPhase): return False
        if self.index != other.index: return False
        if set(self.targets) == set(other.targets) and self.phase == other.phase:
            return True
        return False

    def __str__(self) -> str:
        return "ParityPhase({!s}, {!s})".format(self.phase, ", ".join(str(t) for t in self.targets))

    def _max_target(self) -> int:
        return max(self.targets)

    def reposition(self, mask, bit_mask = None):
        g = self.copy()
        g.targets = [mask[t] for t in g.targets]
        return g

    def to_basic_gates(self):
        cnots = [CNOT(self.targets[i],self.targets[i+1]) for i in range(len(self.targets)-1)]
        p = ZPhase(self.targets[-1], self.phase)
        return cnots + [p] + list(reversed(cnots))

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

    def tcount(self):
        return 1 if self.phase.denominator > 2 else 0


class RZZ(ParityPhase):
    name = 'RZZ'
    qasm_name = 'rzz'
    def __init__(self, control: int, target: int, phase: FractionLike) -> None:
        self.control = control
        self.target = target
        self.targets = (control, target)
        self.phase = phase


class FSim(Gate):
    name = 'FSim'
    qsim_name = 'fs'
    print_phase = True
    def __init__(self, control:int, target:int, theta:FractionLike, phi:FractionLike):
        # TODO: this version assumes theta is always (pi/2)
        assert theta == Fraction(1, 2)
        self.control = control
        self.target = target
        self.theta = theta
        self.phi = phi
        self.phases = [theta, phi]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, FSim): return False
        if self.index != other.index: return False
        if (self.control == other.control and self.target == other.target and
            self.theta == other.theta and self.phi == other.phi):
            return True
        return False

    def __str__(self) -> str:
        return "FSim({!s}, {!s}, {!s}, {!s})".format(self.control, self.target, self.theta, self.phi)

    def reposition(self, mask, bit_mask = None):
        g = self.copy()
        g.targets = [mask[t] for t in g.targets]
        return g

    def to_basic_gates(self):
        # TODO
        #cnots = [CNOT(self.targets[i],self.targets[i+1]) for i in range(len(self.targets)-1)]
        #p = ZPhase(self.targets[-1], self.phase)
        return [self]

    def to_graph(self, g, q_mapper, _c_mapper):
        # TODO: this version assumes theta is always (pi/2)
        r = max(q_mapper.next_row(self.target), q_mapper.next_row(self.control))
        qmin = min(self.target,self.control)
        c0 = self.graph_add_node(g, q_mapper, VertexType.Z, self.control,r)
        t0 = self.graph_add_node(g, q_mapper, VertexType.Z, self.target,r)
        c = g.add_vertex(VertexType.Z, self.control, r+1)
        t = g.add_vertex(VertexType.Z, self.target, r+1)
        g.add_edge((c0, t))
        g.add_edge((t0, c))
        q_mapper.set_prev_vertex(self.control, c)
        q_mapper.set_prev_vertex(self.target, t)

        pg0 = g.add_vertex(VertexType.Z, qmin+0.5,r+2)
        pg1 = g.add_vertex(VertexType.Z, qmin+0.5,r+3)

        g.set_phase(c, Fraction(-1,2) * self.phi)
        g.set_phase(t, Fraction(-1,2) * self.phi)
        g.set_phase(pg1, (Fraction(1,2) * self.phi) - Fraction(1,2))

        g.add_edge((c, pg0), EdgeType.HADAMARD)
        g.add_edge((t, pg0), EdgeType.HADAMARD)
        g.add_edge((pg0, pg1), EdgeType.HADAMARD)

        #rs[self.target] = r+2
        #rs[self.control] = r+2

        q_mapper.shift_all_rows(4)
        g.scalar.add_power(1)

        # c1 = self.graph_add_node(g, q_mapper, VertexType.Z, self.control,r)
        # t1 = self.graph_add_node(g, q_mapper, VertexType.Z, self.target,r)
        # c2 = self.graph_add_node(g, q_mapper, VertexType.Z, self.control,r+1)
        # t2 = self.graph_add_node(g, q_mapper, VertexType.Z, self.target,r+1)

        # pg1 = g.add_vertex(VertexType.Z,qmin-0.5,r+1)
        # pg1b = g.add_vertex(VertexType.Z,qmin-1.5,r+1, phase=self.theta)
        # g.add_edge((c1,pg1),EdgeType.HADAMARD)
        # g.add_edge((t1,pg1),EdgeType.HADAMARD)
        # g.add_edge((pg1,pg1b),EdgeType.HADAMARD)

        # pg2 = g.add_vertex(VertexType.Z,qmin-0.5,r+2)
        # pg2b = g.add_vertex(VertexType.Z,qmin-1.5,r+2, phase=-self.theta)
        # g.add_edge((c1,pg2),EdgeType.HADAMARD)
        # g.add_edge((t1,pg2),EdgeType.HADAMARD)
        # g.add_edge((c2,pg2),EdgeType.HADAMARD)
        # g.add_edge((t2,pg2),EdgeType.HADAMARD)
        # g.add_edge((pg2,pg2b),EdgeType.HADAMARD)

        # if zh_form:
        #     hbox = g.add_vertex(VertexType.H_BOX,qmin-0.5,r+3, phase=self.phi)
        #     g.add_edge((c2,hbox),EdgeType.SIMPLE)
        #     g.add_edge((t2,hbox),EdgeType.SIMPLE)
        # else:
        #     half_phi = Fraction(1,2) * self.phi
        #     pg3 = g.add_vertex(VertexType.Z,qmin-0.5,r+3)
        #     pg3b = g.add_vertex(VertexType.Z,qmin-1.5,r+3, phase=half_phi)
        #     g.add_edge((c2,pg3),EdgeType.HADAMARD)
        #     g.add_edge((t2,pg3),EdgeType.HADAMARD)
        #     g.add_edge((pg3,pg3b),EdgeType.HADAMARD)
        #     g.set_phase(c2, -half_phi)
        #     g.set_phase(t2, -half_phi)

        # q_mapper.shift_all_rows(5)

        #h = g.add_vertex(VertexType.H_BOX, qmin + 0.5, r + 0.5)
        #g.add_edge((t,h),EdgeType.SIMPLE)
        #g.add_edge((c1,h),EdgeType.SIMPLE)
        #g.add_edge((c2,h),EdgeType.SIMPLE)

        #rs[self.target] = r+4
        #rs[self.control] = r+4

        # TODO
        #for gate in self.to_basic_gates():
        #    gate.to_graph(g, q_mapper, c_mapper)

    def tcount(self):
        # TODO
        return 0 #1 if self.phase.denominator > 2 else 0

class CCZ(Gate):
    name = 'CCZ'
    qasm_name = 'ccz'
    qc_name = 'Z'
    quipper_name = 'Z'
    def __init__(self, ctrl1: int, ctrl2: int, target: int):
        self.target = target
        self.ctrl1 = ctrl1
        self.ctrl2 = ctrl2
    def __str__(self) -> str:
        return "{}(c1={!s},c2={!s},t={!s})".format(self.name,self.ctrl1,self.ctrl2,self.target)
    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CCZ): return False
        if self.index != other.index: return False
        if set([self.target,self.ctrl1,self.ctrl2]) == set([other.target,other.ctrl1,other.ctrl2]):
            return True
        return False

    def _max_target(self):
        return max([self.target,self.ctrl1,self.ctrl2])

    def tcount(self):
        return 7

    def reposition(self, mask, bit_mask = None):
        g = self.copy()
        g.target = mask[g.target]
        g.ctrl1 = mask[g.ctrl1]
        g.ctrl2 = mask[g.ctrl2]
        return g

    def to_basic_gates(self):
        c1,c2,t = self.ctrl1, self.ctrl2, self.target
        return [CNOT(c2,t), T(t,adjoint=True),
                CNOT(c1,t), T(t),CNOT(c2,t),T(t,adjoint=True),
                CNOT(c1,t), T(c2), T(t), CNOT(c1,c2),T(c1),
                T(c2,adjoint=True),CNOT(c1,c2)]

    def to_graph(self, g, q_mapper, _c_mapper):
        # if basic_gates:
        #     for gate in self.to_basic_gates():
        #         gate.to_graph(g, q_mapper, c_mapper)
        # else:
        r = max(q_mapper.next_row(self.target), q_mapper.next_row(self.ctrl1), q_mapper.next_row(self.ctrl2))
        qmin = min(self.target,self.ctrl1,self.ctrl2)
        t = self.graph_add_node(g, q_mapper, VertexType.Z, self.target, r)
        c1 = self.graph_add_node(g, q_mapper, VertexType.Z, self.ctrl1, r)
        c2 = self.graph_add_node(g, q_mapper, VertexType.Z, self.ctrl2, r)
        h = g.add_vertex(VertexType.H_BOX, qmin + 0.5, r + 0.5)
        g.add_edge((t,h),EdgeType.SIMPLE)
        g.add_edge((c1,h),EdgeType.SIMPLE)
        g.add_edge((c2,h),EdgeType.SIMPLE)
        q_mapper.set_next_row(self.target, r+1)
        q_mapper.set_next_row(self.ctrl1, r+1)
        q_mapper.set_next_row(self.ctrl2, r+1)

    def to_quipper(self):
        s = 'QGate["{}"]({!s})'.format(self.quipper_name,self.target)
        s += ' with controls=[+{!s},+{!s}]'.format(self.ctrl1,self.ctrl2)
        s += ' with nocontrol'
        return s

class Tofolli(CCZ):
    name = 'Tof'
    qasm_name = 'ccx'
    qc_name = 'Tof'
    quipper_name = 'not'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Tofolli): return False
        if self.index != other.index: return False
        if self.target != other.target: return False
        if set([self.ctrl1,self.ctrl2]) == set([other.ctrl1,other.ctrl2]):
            return True
        return False

    def to_basic_gates(self):
        c1,c2,t = self.ctrl1, self.ctrl2, self.target
        return [HAD(t), CNOT(c2,t), T(t,adjoint=True),
                CNOT(c1,t), T(t),CNOT(c2,t),T(t,adjoint=True),
                CNOT(c1,t), T(c2), T(t), CNOT(c1,c2),T(c1),
                T(c2,adjoint=True),CNOT(c1,c2), HAD(t)]

    def to_graph(self, g, q_mapper, c_mapper):
        t = self.target
        HAD(t).to_graph(g, q_mapper, c_mapper)
        CCZ.to_graph(self, g, q_mapper, c_mapper)
        HAD(t).to_graph(g, q_mapper, c_mapper)

class CSWAP(CCZ):
    name = 'CSWAP'
    qasm_name = 'cswap'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CSWAP): return False
        if self.index != other.index: return False
        if self.ctrl1 != other.ctrl1: return False
        if set([self.ctrl2,self.target]) == set([other.ctrl2,other.target]):
            return True
        return False

    def to_basic_gates(self):
        c, t1, t2 = self.ctrl1, self.ctrl2, self.target
        return [CNOT(control=t2,target=t1)] + \
               Tofolli(c,t1,t2).to_basic_gates() + \
               [CNOT(control=t2,target=t1)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class U2(Gate):  # See https://arxiv.org/pdf/1707.03429.pdf
    name = 'U2'
    qasm_name = 'u2'
    print_phase = True
    def __init__(self, target: int, theta: FractionLike, phi: FractionLike) -> None:
        self.target = target
        self.theta = theta
        self.phi = phi
        self.phases = [theta, phi]

    def to_basic_gates(self):
        return [ZPhase(self.target,phase=(self.phi-Fraction(1,2))%2),
                XPhase(self.target,phase=Fraction(1,2)),
                ZPhase(self.target,phase=(self.theta+Fraction(1,2))%2)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class U3(Gate):  # See equation (5) of https://arxiv.org/pdf/1707.03429.pdf
    name = 'U3'
    qasm_name = 'u3'
    print_phase = True
    def __init__(self, target: int, theta: FractionLike, phi: FractionLike, rho: FractionLike) -> None:
        self.target = target
        self.theta = theta
        self.phi = phi
        self.rho = rho
        self.phases = [theta, phi, rho]

    def to_basic_gates(self):
        return [ZPhase(self.target,phase=self.rho),
                XPhase(self.target,phase=Fraction(1,2)),
                ZPhase(self.target,phase=(self.theta+1)%2),
                XPhase(self.target,phase=Fraction(1,2)),
                ZPhase(self.target,phase=(self.phi+3)%2)]

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class CU3(Gate):
    name = 'CU3'
    qasm_name = 'cu3'
    print_phase = True

    def __init__(self, control: int, target: int, theta: FractionLike, phi: FractionLike, rho: FractionLike) -> None:
        self.control = control
        self.target = target
        self.theta = theta
        self.phi = phi
        self.rho = rho
        self.phases = [theta, phi, rho]

    def to_basic_gates(self):
        return [ZPhase(self.control,phase=(self.rho+self.phi)/2),
                ZPhase(self.target, phase=(self.rho-self.phi)/2),
                CNOT(self.control,self.target)] + \
               U3(self.target,-self.theta/2,0,-(self.phi+self.rho)/2).to_basic_gates() + \
               [CNOT(self.control, self.target)] + \
               U3(self.target,self.theta/2,self.phi,0).to_basic_gates()

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)


class CU(Gate):
    name = 'CU'
    qasm_name = 'cu'
    print_phase = True

    def __init__(self, control: int, target: int, theta: FractionLike, phi: FractionLike, rho: FractionLike,
            gamma: FractionLike) -> None:
        self.control = control
        self.target = target
        self.theta = theta
        self.phi = phi
        self.rho = rho
        self.gamma = gamma
        self.phases = [theta, phi, rho, gamma]

    def to_basic_gates(self):
        return [ZPhase(self.control,phase=self.gamma)] + \
               CU3(self.control,self.target,self.theta,self.phi,self.rho).to_basic_gates()

    def to_graph(self, g, q_mapper, c_mapper):
        for gate in self.to_basic_gates():
            gate.to_graph(g, q_mapper, c_mapper)

class InitAncilla(Gate):
    name = 'InitAncilla'
    def __init__(self, label):
        self.label = label
        self.target = label

class PostSelect(Gate):
    name = 'PostSelect'
    def __init__(self, label):
        self.label = label
        self.target = label

    def to_graph(self, g, labels, qs, _cs, rs, _crs):
        v = g.add_vertex(VertexType.Z, self.label, 0)

class DiscardBit(Gate):
    name = 'DiscardBit'
    def __init__(self, target):
        self.target = target

    def reposition(self, _mask, bit_mask = None):
        g = self.copy()
        g.target = bit_mask[g.target]
        return g

    def to_graph(self, g, _q_mapper, c_mapper):
        r = c_mapper.next_row(self.target)
        self.graph_add_node(g,
            c_mapper,
            VertexType.Z,
            self.target,
            r,
            ground=True)
        u = g.add_vertex(VertexType.X, c_mapper.to_qubit(self.target), r+1)
        c_mapper.set_prev_vertex(self.target, u)
        c_mapper.set_next_row(self.target, r+2)

class Measurement(Gate):
    target: int
    result_bit: Optional[int]

    name = 'Measurement'
    quipper_name = 'measure'
    # This gate has special syntax in qasm: https://openqasm.com/language/insts.html#measurement
    # PyZX supports the following subset of the syntax:
    # * (OpenQASM 2) measure q[0] -> c[0]
    # * (OpenQASM 3) c[0] = measure q[0]

    def __init__(self, target: int, result_bit: Optional[int] = None) -> None:
        self.target = target
        self.result_bit = result_bit

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Measurement): return False
        if self.target != other.target: return False
        if self.result_bit != other.result_bit: return False
        return True

    def reposition(self, mask, bit_mask = None):
        g = self.copy()
        g.target = mask[self.target]
        if self.result_bit is not None and bit_mask is not None:
            g.result_bit = bit_mask[self.result_bit]
        return g

    def to_graph(self, g, q_mapper, c_mapper):
        # Discard previous bit value
        if self.result_bit is not None:
            DiscardBit(self.result_bit).to_graph(g, q_mapper, c_mapper)
        # Qubit measurement
        r = q_mapper.next_row(self.target)
        if self.result_bit is not None:
            r = max(r, c_mapper.next_row(self.result_bit))
        v = self.graph_add_node(g,
            q_mapper,
            VertexType.Z,
            self.target,
            r,
            ground=True)
        q_mapper.set_next_row(self.target, r+1)
        # Classical result
        if self.result_bit is not None:
            u = self.graph_add_node(g,
                c_mapper,
                VertexType.X,
                self.result_bit,
                r)
            g.add_edge((v,u), EdgeType.SIMPLE)
            c_mapper.set_next_row(self.result_bit, r+1)

gate_types: Dict[str,Type[Gate]] = {
    "XPhase": XPhase,
    "NOT": NOT,
    "SX": SX,
    "CSX": CSX,
    "YPhase": YPhase,
    "Y": Y,
    "CY": CY,
    "ZPhase": ZPhase,
    "CPhase": CPhase,
    "CP": CPhase,
    "Z": Z,
    "S": S,
    "T": T,
    "CNOT": CNOT,
    "CZ": CZ,
    "ParityPhase": ParityPhase,
    "XCX": XCX,
    "SWAP": SWAP,
    "CSWAP": CSWAP,
    "CRZ": CRZ,
    "HAD": HAD,
    "H": HAD,
    "CHAD": CHAD,
    "TOF": Tofolli,
    "CCZ": CCZ,
    "U2": U2,
    "U3": U3,
    "U": U3,
    "CU3": CU3,
    "CU": CU,
    "CRX": CRX,
    "CRY": CRY,
    "RZZ": RZZ,
    "RXX": RXX,
    "FSim": FSim,
    "InitAncilla": InitAncilla,
    "PostSelect": PostSelect,
    "DiscardBit": DiscardBit,
    "Measurement": Measurement,
}

qasm_gate_table: Dict[str, Type[Gate]] = {
    "x": NOT,
    "y": Y,
    "z": Z,
    "s": S,
    "t": T,
    "sdg": S,
    "tdg": T,
    "sx": SX,
    "sxdg": SX,
    "csx": CSX,
    "h": HAD,
    "rx": XPhase,
    "ry": YPhase,
    "rz": ZPhase,
    "cp": CPhase,
    "cphase": CPhase,
    "cu1": CPhase,
    "p": ZPhase,
    "u1": ZPhase,
    "u2": U2,
    "u3": U3,
    "u": U3,
    "U": U3,  # needed for OpenQASM 3 in Qiskit 1.0 and up
    "cu3": CU3,
    "cu": CU,
    "cx": CNOT,
    "CX": CNOT,  # needed for backwards compatibility with older versions of qiskit
    "cy": CY,
    "cz": CZ,
    "ch": CHAD,
    "crx": CRX,
    "cry": CRY,
    "crz": CRZ,
    "rxx": RXX,
    "rzz": RZZ,
    "ccx": Tofolli,
    "ccz": CCZ,
    "swap": SWAP,
    "cswap": CSWAP,
    "measure": Measurement,
}



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Tuple, cast
from typing_extensions import Literal

from .circuit.gates import CNOT

def parity_network(n: int, S: List[List[Literal[0,1]]]) -> List[CNOT]:
    """Implements the parity network algorithm of https://arxiv.org/pdf/1712.01859.pdf.
    Specifically, it follows the pseudo-code of page 14."""
    c = [] # List of cnots
    Q: List[Tuple[List[List[Literal[0,1]]],List[int],int]] = [] # stack
    Q.append((S,list(range(n)),-1))
    while Q:
        S, I, i = Q.pop()
        if not S or not I: continue
        if i != -1:
            while True:
                for j in range(n):
                    if j==i: continue
                    if all(y[j] for y in S):
                        c.append(CNOT(j,i))
                        for (Sp,Ip,ip) in (Q+[(S,I,i)]):
                            for y in Sp:
                                y[j] = cast(Literal[0,1], (y[i]+y[j])%2)
                        break
                else:
                    break
        j = max(I, key=lambda j: max([len([y for y in S if y[j]==0]),len([y for y in S if y[j]==1])]))
        S0 = [y.copy() for y in S if y[j]==0]
        S1 = [y.copy() for y in S if y[j]==1]
        Iprime = [jp for jp in I if jp!=j]
        if i == -1:
            Q.append((S1,Iprime,j))
        else:
            Q.append((S1,[jp for jp in I if jp!=i],i))
        Q.append((S0,Iprime, i))
    return c



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from fractions import Fraction
import unittest
import sys
import random
from types import ModuleType
from typing import Optional

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.circuit import Circuit
from pyzx_param.graph import Graph, EdgeType, Scalar, VertexType
from pyzx_param.simulate import (
    replace_magic_states,
    cut_vertex,
    cut_edge,
    gen_catlike_term,
    replace_u3_states
)
from pyzx_param.generate import cliffords
from pyzx_param.simplify import full_reduce

np: Optional[ModuleType]
try:
    import numpy as np
except ImportError:
    np = None

def rand_graph(qubits=5,depth=10):
    g = cliffords(qubits,depth)
    g.apply_state('0'*qubits)
    g.apply_effect('0'*qubits)
    return g
    
def round_complex(scalar,decimal_places):
    return round(scalar.real,decimal_places) + round(scalar.imag,decimal_places)*1j

@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestSimulate(unittest.TestCase):

    def test_magic_state_decomposition_is_correct(self):
        c = Circuit(6)
        for i in range(6):
            c.add_gate("T", i)
        g = c.to_graph()
        gsum = replace_magic_states(g)
        self.assertTrue(np.allclose(g.to_tensor(), gsum.to_tensor()))
        
    def test_vertex_cut(self,repeats=20):
        for i in range(1,repeats):
            g = rand_graph() # generate random Clifford graph
            v_cut = random.randrange(len(g.vertices()))
            g0,g1 = cut_vertex(g,v_cut) # apply random vertex cut
            
            for g_i in (g,g0,g1): full_reduce(g_i)
            
            scal    = round_complex(g.scalar.to_number(),3) # the scalar from fully reducing g
            scalCut = round_complex(g0.scalar.to_number()+g1.scalar.to_number(),3) # the sum of scalars from the cut graph
            assert(scal == scalCut)
        
    def test_edge_cut(self,repeats=20):
        for i in range(1,repeats):
            g = rand_graph() # generate random Clifford graph
            rand_v = random.randrange(len(g.vertices()))
            rand_neigh = list(g.neighbors(rand_v))[random.randrange(len(g.neighbors(rand_v)))]
            e_cut = (rand_v,rand_neigh)  # apply random edge cut
            
            g0,g1 = cut_edge(g,e_cut)
            for g_i in (g,g0,g1): full_reduce(g_i)
            
            scal    = round_complex(g.scalar.to_number(),3) # the scalar from fully reducing g
            scalCut = round_complex(g0.scalar.to_number()+g1.scalar.to_number(),3) # the sum of scalars from the cut graph
            assert(scal == scalCut)

    def test_cat_decomp_scalar(self) -> None:
        """Test the scalars of generated cat-like terms are correct."""
        g = Graph()

        # Generate random scalar parameters
        phase = Fraction(random.randint(1, 10), random.randint(1, 10))
        power = random.randint(1, 10)
        positive = bool(random.randint(0, 1))

        # Generate cat-like term with random scalar parameters
        G = gen_catlike_term(g, [],
                             Fraction(1, 1),
                             Fraction(1, 1),
                             Fraction(1, 1),
                             EdgeType.SIMPLE,
                             EdgeType.SIMPLE,
                             positive,
                             power,
                             phase,
                             False)

        # Create expected scalar
        s = Scalar()
        s.add_power(power)
        s.add_phase(phase)
        if not positive:
            s.add_phase(1)

        # Check if the scalar from generated term is correct
        self.assertTrue(G.scalar.to_number() == s.to_number())

    def test_replace_arbitrary_rotation(self):
        theta = Fraction(1, 7)
        g = Graph()
        v0 = g.add_vertex(VertexType.Z, qubit=0, row=0, phase=theta)
        o1 = g.add_vertex(VertexType.BOUNDARY, qubit=0, row=1)
        g.add_edge((o1, v0), EdgeType.HADAMARD)

        g.set_outputs((o1,))
        t1 = g.to_tensor()

        gsum = replace_u3_states(g)
        assert len(gsum.graphs) == 2
        t2 = sum([g_.to_tensor() for g_ in gsum.graphs])

        self.assertTrue(np.allclose(t1, t2))


    def test_replace_rotation_pair(self):
        theta = Fraction(1, 7)
        g = Graph()
        v0 = g.add_vertex(VertexType.Z, qubit=0, row=0, phase=theta)
        v1 = g.add_vertex(VertexType.Z, qubit=1, row=0, phase=-theta)
        o1 = g.add_vertex(VertexType.BOUNDARY, qubit=0, row=1)
        o2 = g.add_vertex(VertexType.BOUNDARY, qubit=1, row=1)
        g.add_edge((o1, v0), EdgeType.HADAMARD)
        g.add_edge((o2, v1), EdgeType.HADAMARD)

        g.set_outputs((o1, o2))
        t1 = g.to_tensor()

        gsum = replace_u3_states(g)
        self.assertTrue(len(gsum.graphs) == 3)
        t2 = sum([g_.to_tensor() for g_ in gsum.graphs])

        self.assertTrue(np.allclose(t1, t2))


    def test_magic_theta_cat_replace(self):
        theta = Fraction(1, 7)
        g = Graph()
        vo = [g.add_vertex(VertexType.BOUNDARY, qubit=i, row=1) for i in range(5)]
        vs = [g.add_vertex(VertexType.Z, qubit=i, row=0, phase=theta) for i in range(5)]
        [g.add_edge((a, b)) for a, b in zip(vo, vs)]
        g.set_outputs(tuple(vo))

        t1 = g.to_tensor()

        gsum = replace_u3_states(g)
        self.assertTrue(len(gsum.graphs) == 4)
        t2 = sum([g_.to_tensor() for g_ in gsum.graphs])

        self.assertTrue(np.allclose(t1, t2))


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import random
import sys
from types import ModuleType
from typing import Optional
from fractions import Fraction

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.circuit import Circuit
from pyzx_param.circuit.gates import CNOT
from pyzx_param.generate import cliffordT
from pyzx_param.simplify import clifford_simp
from pyzx_param.extract import extract_circuit
from pyzx_param import simplify

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import tensorfy, compare_tensors
except ImportError:
    np = None


SEED = 1337


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
class TestExtract(unittest.TestCase):

    def test_simple_extract(self):
        c = Circuit(1)
        c.add_gate("HAD", 0)
        c.add_gate("ZPhase", 0, phase= Fraction(1,4))

        g = c.to_graph()
        
        simplify.full_reduce(g,quiet=True)

        c2 = extract_circuit(g)
        self.assertListEqual(c.gates, c2.gates)
        self.assertTrue(c.verify_equality(c2))

    def test_extract_not_graph_like(self):
        c = Circuit(1)
        c.add_gate("HAD", 0)
        c.add_gate("ZPhase", 0, phase=1/4)

        g = c.to_graph()
        

        with self.assertRaises(ValueError) as context:
            c = extract_circuit(g)
        self.assertTrue("Input graph is not graph-like. Try running full_reduce first" in str(context.exception))



    def test_extract_circuit(self):
        random.seed(SEED)
        for i in range(5):
            circ = cliffordT(4,50,0.1)
            t = tensorfy(circ,False)
            clifford_simp(circ,quiet=True)
            with self.subTest(i=i):
                c = extract_circuit(circ)
                t2 = c.to_tensor(False)
                self.assertTrue(compare_tensors(t,t2,False))

    def test_cz_optimize_extract(self):
        qb_no = 8
        c = Circuit(qb_no)
        for i in range(qb_no):
            for j in range(i+1,qb_no):
                c.add_gate("CZ",i,j)

        g = c.to_graph()
        clifford_simp(g,quiet=True)
        c2 = extract_circuit(g)
        cnot_count = 0
        for gate in c2.gates:
            if isinstance(gate, CNOT):
                cnot_count+=1
        self.assertTrue(cnot_count==4)
        self.assertTrue(c.verify_equality(c2))
        


if __name__ == '__main__':
    unittest.main()


import numpy as np

from enum import Enum
from typing import List, Optional, Tuple

from pyzx_param.circuit import Circuit

from ..linalg import Mat2
from .architecture import (
    Architecture,
    create_fully_connected_architecture,
)
from .parity_maps import CNOT_tracker
from .machine_learning import GeneticAlgorithm, ParticleSwarmOptimization

from .steiner import rec_steiner_gauss as steiner_gauss


class ElimMode(Enum):
    """
    Row elimination modes for the cnot mapper procedures
    """

    GAUSS_MODE = "gauss"
    """Gaussian elimination, ignoring the architecture."""

    STEINER_MODE = "steiner"
    """Steiner tree based Gaussian elimination, optimizing the number of SWAPs
    operations required to synthesize the CNOTs on the restricted
    architecture."""

    GENETIC_GAUSS_MODE = "genetic_gauss"
    """Gauss elimination using a genetic algorithm to find the best row permutation."""

    GENETIC_STEINER_MODE = "genetic_steiner"
    """Steiner Gauss elimination using a genetic algorithm to find the best row permutation."""

    PSO_GAUSS_MODE = "pso_gauss"
    """Gauss elimination using Particle Swarm Optimization to find the best row permutation."""

    PSO_STEINER_MODE = "pso_steiner"
    """Steiner Gauss elimination using Particle Swarm Optimization to find the best row permutation."""

    def __str__(self):
        return f"{self.value}"


genetic_elim_modes = [ElimMode.GENETIC_STEINER_MODE, ElimMode.GENETIC_GAUSS_MODE]
pso_elim_modes = [ElimMode.PSO_GAUSS_MODE, ElimMode.PSO_STEINER_MODE]
basic_elim_modes = [ElimMode.STEINER_MODE, ElimMode.GAUSS_MODE]
elim_modes = list(ElimMode)


class CostMetric(Enum):
    """
    Metrics for the cost of the gates needed for a given permutation,
    used by the cnot mapper fitness functions.
    """

    COMBINED = "combined"
    """Count both the number of CNOTs and the depth of the circuit"""
    DEPTH = "depth"
    """Count the number of CNOTs in the circuit"""
    COUNT = "count"
    """Count the depth of the circuit"""

    def __str__(self):
        return f"{self.value}"


class FitnessFunction(object):
    """
    A fitness function that calculates the cost of the gates needed for a given permutation.
    """

    def __init__(
        self,
        metric: CostMetric,
        matrix: Mat2,
        mode: Optional[ElimMode],
        architecture: Optional[Architecture],
        row: bool = True,
        col: bool = True,
        full_reduce: bool = True,
        **kwargs,
    ):
        """
        Creates and returns a fitness function using the given metric.

        :param metric_func: The metric to use for the fitness function
        :param mode: The type of Gaussian elimination to be used
        :param matrix: A Mat2 parity map to route.
        :param architecture: The architecture to take into account when routing
        :param row: Whether to find a row permutation
        :param col: Whether to find a column permutation
        :param full_reduce: Whether to fully reduce the matrix, thus rebuild the full circuit.
        """
        self.metric = metric
        self.matrix = matrix
        self.mode = mode
        self.architecture = architecture
        self.row = row
        self.col = col
        self.full_reduce = full_reduce
        self.n_qubits = architecture.n_qubits if architecture else matrix.cols()
        self.kwargs = kwargs

    def _make_function(self):
        if self.metric == CostMetric.COMBINED:
            f = lambda c: c.cnot_depth() * 10000 + c.count_cnots()
        elif self.metric == CostMetric.COUNT:
            f = lambda c: c.count_cnots()
        elif self.metric == CostMetric.DEPTH:
            f = lambda c: c.cnot_depth()

        def fitness_func(permutation):
            row_perm = permutation if self.row else np.arange(len(self.matrix.data))
            col_perm = permutation if self.col else np.arange(len(self.matrix.data[0]))
            circuit = CNOT_tracker(self.n_qubits)
            mat = Mat2([[self.matrix.data[r][c] for c in col_perm] for r in row_perm])
            gauss(
                self.mode,
                mat,
                architecture=self.architecture,
                y=circuit,
                full_reduce=self.full_reduce,
                **self.kwargs,
            )
            return f(circuit)

        return fitness_func

    def __call__(self, permutation):
        f = self._make_function()
        return f(permutation)


def gauss(
    mode: Optional[ElimMode],
    matrix: Mat2,
    architecture: Optional[Architecture] = None,
    permutation: Optional[List[int]] = None,
    try_transpose: bool = False,
    **kwargs,
) -> int:
    """
    Performs architecture-aware Gaussian Elimination on a matrix.

    :param mode: Type of Gaussian elimination to be used, see :class:`ElimMode`.
    :param matrix: Target matrix to be reduced.
    :param architecture: Device architecture to take into account.
    :param permutation: If given, reduce a permuted version of the matrix.
    :param kwargs: Other arguments that can be given to the :meth:`Mat2.gauss` function or parameters for the genetic algorithm.
    :return: The rank of the matrix. :data:`matrix` is transformed inplace.
    """
    if try_transpose:
        matrix = matrix.transpose()
        if architecture is not None:
            architecture = architecture.transpose()
    if mode is None:
        mode = ElimMode.GAUSS_MODE

    if mode == ElimMode.GAUSS_MODE:
        # TODO - adjust to get the right gate locations for the given permutation.

        if permutation is not None:
            # print("\033[91m Warning: Permutation parameter with Gauss-Jordan elimination is not yet supported, it can be optimized with permuted_gauss(). \033[0m ")
            # return matrix.gauss(**kwargs)
            # Broken code that tries to implement this.
            matrix = Mat2([[row[i] for i in permutation] for row in matrix.data])
            old_x, old_y = None, None
            if "x" in kwargs:
                old_x = kwargs["x"]
            if "y" in kwargs:
                old_y = kwargs["y"]
            n_qubits = len(matrix.data)
            x = CNOT_tracker(n_qubits)
            kwargs["x"] = x
            kwargs["y"] = None
            rank = matrix.gauss(**kwargs)
            # for gate in x.gates:
            #    #c = permutation[gate.control]
            #    #t = permutation[gate.target]
            #    if old_x != None: old_x.row_add(c, t)
            #    if old_y != None: old_y.col_add(t, c)
            # return rank
        else:
            rank = matrix.gauss(**kwargs)
    elif mode == ElimMode.STEINER_MODE:
        if architecture is None:
            print(
                f"\033[91m Warning: Architecture is not given, assuming fully connected architecture of size {matrix.rows()}. \033[0m "
            )
            architecture = create_fully_connected_architecture(matrix.rows())
        if permutation is not None:
            matrix.permute_rows(permutation)
            matrix.permute_cols(permutation)
        steiner_gauss(matrix, architecture, **kwargs)
        try:
            len(matrix.data)
        except:
            print(
                "Failing length for matrix data:",
                matrix.data,
                "with type",
                type(matrix.data),
            )
        rank = 0  # TODO: The rank is not being computed here
    elif mode == ElimMode.GENETIC_STEINER_MODE:
        perm, circuit, rank = permuted_gauss(
            matrix,
            ElimMode.STEINER_MODE,
            architecture=architecture,
            permutation=permutation,
            **kwargs,
        )
        # return rank
    elif mode == ElimMode.GENETIC_GAUSS_MODE:
        perm, circuit, rank = permuted_gauss(
            matrix,
            ElimMode.GAUSS_MODE,
            architecture=architecture,
            permutation=permutation,
            **kwargs,
        )
    else:
        raise KeyError(f"Invalid elimination mode '{mode}'")
    if try_transpose:
        # TODO - fix x and y circuits... - Needed?
        # TODO pick which gauss version was chosen
        pass
    return rank


def permuted_gauss(
    matrix: Mat2,
    mode: Optional[ElimMode] = None,
    architecture: Optional[Architecture] = None,
    population_size: int = 30,
    crossover_prob: float = 0.8,
    mutate_prob: float = 0.2,
    n_iterations: int = 5,
    row: bool = True,
    col: bool = True,
    full_reduce: bool = True,
    fitness_func: Optional[FitnessFunction] = None,
    x=None,
    y=None,
    **kwargs,
) -> Tuple[List[int], Circuit, int]:
    """
    Applies gaussian elimination to the given matrix, finding an optimal
    permutation of the matrix to reduce the number of CNOT gates.

    :param matrix: Mat2 matrix to do gaussian elimination over
    :param mode: Elimination mode to use
    :param architecture: Architecture to take into account
    :param population_size: For the genetic algorithm
    :param crossover_prob: For the genetic algorithm
    :param mutate_prob: For the genetic algorithm
    :param n_iterations: For the genetic algorithm
    :param row: If the rows should be permutedA
    :param col: If the columns should be permuted
    :param full_reduce: Whether to do full gaussian reduction
    :param fitness_func: Optional fitness function to use
    :param x: Optional tracker for the row operations
    :param y: Optional tracker for the column operations
    :return: Best permutation found, list of CNOTS corresponding to the
        elimination.
    """
    if row or col:
        if fitness_func is None:
            fitness_func = FitnessFunction(
                CostMetric.COMBINED,
                matrix,
                mode,
                architecture,
                row=row,
                col=col,
                full_reduce=full_reduce,
                **kwargs,
            )
        optimizer = GeneticAlgorithm(
            population_size,
            crossover_prob,
            mutate_prob,
            fitness_func,
        )
        permsize = len(matrix.data) if row else len(matrix.data[0])
        best_permutation = optimizer.find_optimum(
            permsize, n_iterations, continued=True
        )
    else:
        best_permutation = np.arange(len(matrix.data))

    n_qubits = len(matrix.data)
    row_perm = best_permutation if row else np.arange(len(matrix.data))
    col_perm = best_permutation if col else np.arange(len(matrix.data[0]))
    if y is None:
        circuit = CNOT_tracker(n_qubits)
    else:
        circuit = y
    mat = Mat2([[matrix.data[r][c] for c in col_perm] for r in row_perm])
    circuit.row_perm = row_perm
    circuit.col_perm = col_perm
    rank = gauss(
        mode, mat, architecture, x=x, y=circuit, full_reduce=full_reduce, **kwargs
    )
    best_permutation = list(best_permutation)
    return best_permutation, circuit, rank


def sequential_gauss(
    matrices: List[Mat2],
    mode: Optional[ElimMode] = None,
    architecture: Optional[Architecture] = None,
    fitness_func: Optional[FitnessFunction] = None,
    input_perm: bool = True,
    output_perm: bool = True,
    swarm_size: int = 15,
    n_steps: int = 5,
    s_crossover: float = 0.4,
    p_crossover: float = 0.3,
    pso_mutation: float = 0.2,
    full_reduce: bool = True,
    **kwargs,
) -> Tuple[List[CNOT_tracker], List[List[int]], int]:
    """
    Applies architecture-aware Gaussian elimination to multiple matrices,
    sharing the optimization passes when using ParticleSwarmOptimization modes.

    :param matrix: List of matrices to do gaussian elimination over
    :param mode: Elimination mode to use
    :param architecture: Architecture to take into account
    :param fitness_func: Optional fitness function to use
    :param input_perm: Allow input permutation
    :param output_perm: Whether the location of the output qubits can be
        different for the input location. Qubit locations can be optimized with
        pso.
    :param swarm_size: Swarm size for the swarm optimization.
    :param n_steps: The number of iterations for the particle swarm optimization.
    :param s_crossover: The crossover percentage with the best particle in the swarm for the particle swarm optimizer. Must be between 0.0 and 1.0.
    :param p_crossover: The crossover percentage with the personal best of a particle for the particle swarm optimizer. Must be between 0.0 and 1.0.
    :param pso_mutation: The mutation percentage of a particle for the particle swarm optimizer. Must be between 0.0 and 1.0.
    :param full_reduce: Fully reduce the matrices
    :return: List of CNOT trackers corresponding to the eliminations, list of
        final permutations for each matrix, and the cost of the eliminations.
    """
    n_qubits = len(matrices[0].data)
    kwargs["full_reduce"] = full_reduce
    circuits: List[CNOT_tracker]
    permutations: List[List[int]] = []
    # print(mode)
    # print(*matrices, sep="\n\n")
    if mode in basic_elim_modes or mode is None:
        circuits = [CNOT_tracker(n_qubits) for _ in matrices]
        permutations = [list(range(n_qubits)) for _ in range(len(matrices) + 1)]
        for i, m in enumerate(matrices):
            gauss(mode, m, architecture=architecture, y=circuits[i], **kwargs)
    elif mode in genetic_elim_modes:
        col = input_perm
        if mode == ElimMode.GENETIC_GAUSS_MODE:
            new_mode = ElimMode.GAUSS_MODE
        else:
            new_mode = ElimMode.STEINER_MODE
        row = True
        circuits = []
        permutations = []
        current_perm = list(range(n_qubits))
        if not col:
            permutations.append(
                current_perm
            )  # Add initial permutation if it is not optimized
        for i, m in enumerate(matrices):
            # Adjust matrix according to current input perm.
            m = Mat2([[row[r] for r in current_perm] for row in m.data])
            if i == len(matrices) - 1:
                row = output_perm  # Last permutation is only optimized if the output qubit locations are flexible.
            perm, circuit, _ = permuted_gauss(
                m,
                new_mode,
                architecture=architecture,
                fitness_func=fitness_func,
                row=row,
                col=col,
                **kwargs,
            )
            # if not col and not row:
            #    perm = current_perm
            circuits.append(circuit)  # type: ignore # Store the extracted circuit
            # Update the new permutation
            current_perm = list(perm)
            if col:
                permutations.append(current_perm)  # Add optimized initial permutation
                if not row:
                    current_perm = list(range(n_qubits))
            permutations.append(current_perm)  # Store the obtained permutation
            col = False  # Subsequent initial permutations are determined by the previous output permutation.
        # input("current perm - should be [0..] ")
    else:  # pso modes
        if mode == ElimMode.PSO_STEINER_MODE:
            new_mode = ElimMode.GENETIC_STEINER_MODE
        else:
            new_mode = ElimMode.GENETIC_GAUSS_MODE
        if not input_perm or not output_perm:
            # You cannot do subsequent passes to optimize the permutation, so pso is useless.
            return sequential_gauss(
                matrices,
                new_mode,
                architecture=architecture,
                fitness_func=fitness_func,
                input_perm=input_perm,
                output_perm=output_perm,
                **kwargs,
            )

        step_func = StepFunction(
            matrices, new_mode, architecture, fitness_func, **kwargs
        )
        optimizer = ParticleSwarmOptimization(
            swarm_size=swarm_size,
            step_func=step_func,
            s_best_crossover=s_crossover,
            p_best_crossover=p_crossover,
            mutation=pso_mutation,
        )
        best_solution = optimizer.find_optimum(
            architecture.n_qubits if architecture is not None else n_qubits,
            n_steps,
            quiet=True,
        )
        circuits, permutations = best_solution
    return (
        circuits,
        permutations,
        sum([c.cnot_depth() * 10000 + c.count_cnots() for c in circuits]),
    )

class StepFunction:
    """
    A step function for the PSO algorithm.
    """

    def __init__(self, matrices, mode, architecture, fitness_func, **kwargs):
        """
        Creates and returns a step function.

        :param matrices: List of parity matrices to reduce sequentially
        :param mode: The type of Gaussian elimination to be used
        :param architecture: The architecture to take into account when routing
        :param fitness_func: Fitness function to guide optimisation
        :param **kwargs: Additional arguments passed
        """
        self.matrices = matrices
        self.mode = mode
        self.architecture = architecture
        self.fitness_func = fitness_func
        self.kwargs = kwargs
        self.rev_matrices = [
            Mat2(np.asarray(m.data).T.tolist()) for m in reversed(matrices)
        ]  # Reverse and transpose the parity matrices to create the reversed equivalent sequence

    def __call__(self, initial_perm):
        """
        Evaluates a candidate qubit permutation by running a forward and reverse sequence of architecture-aware Gaussian eliminations.

        :param initial_perm: The initial qubit permutation to evaluate
        :return: The optimised initial permutation, the resulting circuits and permutations, and the total cost score
        """
        matrices = self.matrices
        new_mode = self.mode
        architecture = self.architecture
        fitness_func = self.fitness_func
        rev_matrices = self.rev_matrices
        kwargs = self.kwargs
        # Apply the original qubit placement
        ms = [
            Mat2([[row[i] for i in initial_perm] for row in m.data])
            if j == 0
            else Mat2([r for r in m.data])
            for j, m in enumerate(matrices)
        ]
        # Optimize the sequence
        circs, perms, score = sequential_gauss(
            ms,
            new_mode,
            architecture=architecture,
            fitness_func=fitness_func,
            input_perm=False,
            output_perm=True,
            **kwargs,
        )
        # Resulting permutation is the initial permutation of the reverse pass
        perms[0] = initial_perm
        ms = [
            Mat2([[row[i] for i in perms[-1]] for row in m.data])
            if j == 0
            else Mat2([r for r in m.data])
            for j, m in enumerate(rev_matrices)
        ]
        # Optimize the reverse sequences.
        _, new_perms, _ = sequential_gauss(
            ms,
            new_mode,
            architecture=architecture,
            fitness_func=fitness_func,
            input_perm=False,
            output_perm=True,
            **kwargs,
        )
        # New initial placement is the final placement of the reverse pass.
        # new_perms[0] = perms[-1]
        return new_perms[-1], (circs, perms), score
